####################################################################################################
# pimms/immutable.py
# Simple class decorator for immutable lazily-loading classes.
# By Noah C. Benson

import copy, types, six, inspect
from .util import (getargspec_py27like, qhash)

# An immutable has three important values in the _pimms_immutable_data_ attribute of its class:
# (1) params
#     A hash from param-name to a tuple:
#     (default, transform_fn, arg_lists, check_fns, deps)
#     * default is None if required or the 1-element list [value] where value is the default-value
#     * transform_fn is None if no transform function is provided, otherwise a transformation of
#       the provided value to a possibly-new value that 'interprets' the value
#     * arg_lists is a list of argument-name-tuples for the check_fns
#     * check_fns is a list of functions that should be called with the attributes matching those
#       named in arg_lists in order to check the new value of the particular parameter
#     * deps is a list of the names of values in the immutable that depend on this param
# (2) values
#     A hash from value-name to a tuple:
#     (arg_list, calc_fn, deps)
#     * arg_list is the list of attribute names on which this value depends, and which should be
#       passed to calc_fn as arguments
#     * calc_fn is the function that actually calculates the value given the attributes in arg_list
#     * deps is the list of dependants on this particular value
# (3) checks
#     A hash of tuples, each of which describes one check function:
#     (arg_list, fn)
# (4) consts
#     A hash of tuples, each of which describes one const-value (value with no inputs)
#     (arg_lists, check_fns)
#     Where the arg_lists and check_fns are as in params.

def is_imm(obj):
    '''
    is_imm(obj) yields True if obj is an instance of an immutable class and False otherwise.
    '''
    return hasattr(type(obj), '_pimms_immutable_data_')
def is_imm_type(cls):
    '''
    is_imm_type(cls) yields True if cls is an immutable class and False otherwise.
    '''
    return hasattr(cls, '_pimms_immutable_data_')
def _imm_is_init(imm):
    dd = object.__getattribute__(imm, '__dict__')
    return '_pimms_immutable_is_init' in dd
def _imm_is_trans(imm):
    dd = object.__getattribute__(imm, '__dict__')
    return '_pimms_immutable_is_trans' in dd
def _imm_is_persist(imm):
    dd = object.__getattribute__(imm, '__dict__')
    return '_pimms_immutable_is_init' not in dd and '_pimms_immutable_is_trans' not in dd
def _imm_param_data(imm):
    return type(imm)._pimms_immutable_data_['params']
def _imm_value_data(imm):
    return type(imm)._pimms_immutable_data_['values']
def _imm_check_data(imm):
    return type(imm)._pimms_immutable_data_['checks']
def _imm_const_data(imm):
    return type(imm)._pimms_immutable_data_['consts']
def _imm_clear(imm):
    dd = object.__getattribute__(imm, '__dict__')
    for val in six.iterkeys(_imm_value_data(imm)):
        if val in dd:
            del dd[val]
    return imm
def _imm_check(imm, names=Ellipsis):
    # runs all the checks on all the names
    all_checks = set([])
    params = _imm_param_data(imm)
    consts = _imm_const_data(imm)
    names = list(params.keys()) + list(consts.keys()) if names is Ellipsis                   else \
            [names]                                   if isinstance(names, six.string_types) else \
            names
    for name in names:
        if name in params:
            (_, _, arg_lists, check_fns, _) = params[name]
        elif name in consts:
            (arg_lists, check_fns) = consts[name]
        else:
            raise ValueError('Attempt to check non-existent param named \'%s\'' % name)
        for (arg_list, check_fn) in zip(arg_lists, check_fns):
            all_checks.add((tuple(arg_list), check_fn))
    # Run the checks; if anything fails, we let the exception rise
    for (arg_list, check_fn) in all_checks:
        if not check_fn(*[getattr(imm, arg) for arg in arg_list]):
            raise RuntimeError('Failed parameter-check on values % of func %s' % (args, check_fn))
    # All checks passed!
    return imm
def _imm_hash(imm):
    c = imm.__class__
    return qhash((c.__module__, c.__name__, imm_params(imm)))
def _imm_default_init(self, *args, **kwargs):
    '''
    An immutable's defalt initialization function is to accept any number of dictionaries followed
    by any number of keyword args and to turn them all into the parameters of the immutable that is
    being created.
    '''
    for (k,v) in six.iteritems({k:v for dct in (args + (kwargs,)) for (k,v) in dct}):
        setattr(self, k, v)
def _imm_init_getattribute(self, name):
    '''
    During the initial transient state, getattribute works on params; as soon as a non-param is
    requested, all checks are forced and the getattr switches to standard transient form.
    '''
    values = _imm_value_data(self)
    params = _imm_param_data(self)
    if name in values:
        _imm_init_to_trans(self)
        return getattr(self, name)
    elif name in params:
        dd = object.__getattribute__(self, '__dict__')
        if name in dd: return dd[name]
        else: raise RuntimeError('Required immutable parameter %s requested before set' % name)
    else:
        # if they request a required param before it's set, raise an exception; that's fine
        return object.__getattribute__(self, name)
def _imm_getattribute(self, name):
    '''
    An immutable's getattribute calculates lazy values when not yet cached in the object then adds
    them as attributes.
    '''
    if _imm_is_init(self):
        return _imm_init_getattribute(self, name)
    else:
        dd = object.__getattribute__(self, '__dict__')
        if name == '__dict__': return dd
        curval = dd.get(name, dd)
        if curval is not dd: return dd[name]
        values = _imm_value_data(self)
        if name not in values:
            return object.__getattribute__(self, name)
        (args, memfn, _) = values[name]
        value = memfn(*[getattr(self, arg) for arg in args])
        dd[name] = value
        # if this is a const, it may have checks to run
        if name in _imm_const_data(self):
            # #TODO
            # Note that there's a race condition that eventually needs to be handled here:
            # If dd[name] is set then a check fails, there may have been something that read the
            # improper value in the meantime
            try: _imm_check(self, [name])
            except Exception:
                del dd[name]
                raise
        # if those pass, then we're fine
        return value
def _imm_init_setattr(self, name, value):
    '''
    An immutable's initial setattr allows only param's to be set and does not run checks on the new
    parameters until a full parameter-set has been specified, at which point it runs all checks and
    switches over to a normal setattr and getattr method.
    '''
    params = _imm_param_data(self)
    if name in params:
        tx_fn = params[name][1]
        value = value if tx_fn is None else tx_fn(value)
        # Set the value
        object.__getattribute__(self, '__dict__')[name] = value
        # No checks are run, as we're in initialization mode...
    else:
        raise TypeError(
            'Attempt to change non-parameter \'%s\' of initializing immutable' % name)
def _imm_trans_setattr(self, name, value):
    '''
    An immutable's transient setattr allows params to be set, and runs checks as they are.
    '''
    params = _imm_param_data(self)
    dd = object.__getattribute__(self, '__dict__')
    if name in params:
        (_, tx_fn, arg_lists, check_fns, deps) = params[name]
        value = value if tx_fn is None else tx_fn(value)
        old_deps = {}
        orig_value = dd[name]
        # clear the dependencies before we run the checks; save them in case the checks fail and we
        # go back to how things were...
        for dep in deps:
            if dep in dd:
                old_deps[dep] = dd[dep]
                del dd[dep]
        try:
            dd[name] = value
            for (args, check_fn) in zip(arg_lists, check_fns):
                if not check_fn(*[getattr(self, arg) for arg in args]):
                    raise RuntimeError(
                        ('Changing value of immutable attribute \'%s\'' +
                         ' caused validation failure: %s') % (name, (args, check_fn)))
            # if all the checks ran, we don't return the old deps; they are now invalid
            old_deps = None
        finally:
            if old_deps:
                # in this case, something didn't check-out, so we return the old deps and let the
                # exception ride; we also return the original value of the edited param
                for (dep,val) in six.iteritems(old_deps):
                    dd[dep] = val
                dd[name] = orig_value
    else:
        raise TypeError(
            'Attempt to change non-parameter member \'%s\' of transient immutable' % name)
def _imm_setattr(self, name, value):
    '''
    A persistent immutable's setattr simply does not allow attributes to be set.
    '''
    if _imm_is_persist(self):
        raise TypeError('Attempt to change parameter \'%s\' of non-transient immutable' % name)
    elif _imm_is_trans(self):
        return _imm_trans_setattr(self, name, value)
    else:
        return _imm_init_setattr(self, name, value)
def _imm_trans_delattr(self, name):
    '''
    A transient immutable's delattr allows the object's value-caches to be invalidated; a var that
    is deleted returns to its default-value in a transient immutable, otherwise raises an exception.
    '''
    (params, values) = (_imm_param_data(self), _imm_value_data(self))
    if name in params:
        dflt = params[name][0]
        if dflt is None:
            raise TypeError(
                'Attempt to reset required parameter \'%s\' of immutable' % name)
        setattr(self, name, dflt[0])
    elif name in values:
        dd = object.__getattribute__(self, '__dict__')
        if name in dd:
            del dd[name]
            if name in _imm_const_data(self): _imm_check(imm, [name])
    else:
        raise TypeError('Cannot delete non-value non-param attribute \'%s\' from immutable' % name)
def _imm_delattr(self, name):
    '''
    A persistent immutable's delattr allows the object's value-caches to be invalidated, otherwise
    raises an exception.
    '''
    if _imm_is_persist(self):
        values = _imm_value_data(self)
        if name in values:
            dd = object.__getattribute__(self, '__dict__')
            if name in dd:
                del dd[name]
                if name in _imm_const_data(self): _imm_check(imm, [name])
        else:
            raise TypeError('Attempt to reset parameter \'%s\' of non-transient immutable' % name)
    else:
        return _imm_trans_delattr(self, name)
def _imm_dir(self):
    '''
    An immutable object's dir function should list not only its attributes, but also its un-cached
    lazy values.
    '''
    dir0 = set(dir(self.__class__))
    dir0.update(self.__dict__.keys())
    dir0.update(six.iterkeys(_imm_value_data(self)))
    return sorted(list(dir0))
def _imm_repr(self):
    '''
    The default representation function for an immutable object.
    '''
    return (type(self).__name__
            + ('(' if _imm_is_persist(self) else '*(')
            + ', '.join([k + '=' + str(v) for (k,v) in six.iteritems(imm_params(self))])
            + ')')
def _imm_new(cls):
    '''
    All immutable new classes use a hack to make sure the post-init cleanup occurs.
    '''
    imm = object.__new__(cls)
    # Note that right now imm has a normal setattr method;
    # Give any parameter that has one a default value
    params = cls._pimms_immutable_data_['params']
    for (p,dat) in six.iteritems(params):
        dat = dat[0]
        if dat: object.__setattr__(imm, p, dat[0])
    # Clear any values; they are not allowed yet
    _imm_clear(imm)
    # Note that we are initializing...
    dd = object.__getattribute__(imm, '__dict__')
    dd['_pimms_immutable_is_init'] = True
    # That should do it!
    return imm
def _imm__copy__(self):
    '''
    The default immutable copy operation yields a new instance of the immutable with an identical
    dictionary.
    '''
    if _imm_is_init(self): raise RuntimeError('Cannot copy an initializing immutable')
    dup = _imm_new(type(self))
    sd = object.__getattribute__(self, '__dict__')
    dd = object.__getattribute__(dup,  '__dict__')
    del dd['_pimms_immutable_is_init']
    for (k,v) in six.iteritems(sd):
        dd[k] = v
    return dup
    
def _imm_init_to_trans(imm):
    # changes state from initializing to transient; runs checks along the way
    # if all params are not yet set, this is an error
    params = _imm_param_data(imm)
    dd = object.__getattribute__(imm, '__dict__')
    if not _imm_is_init(imm):
        raise RuntimeError(
            'Attempted to change non-initializing immutable from initializing to transient')
    if not all(p in dd for p in six.iterkeys(params)):
        diffs = [p for p in six.iterkeys(params) if p not in dd]
        raise RuntimeError('Not all parameters were set prior to accessing values: %s' % (diffs,))
    # Okay, we can run the checks now; we need to remove init status, though...
    del dd['_pimms_immutable_is_init']
    dd['_pimms_immutable_is_trans'] = True
    _imm_check(imm)
    # Those passed, so we can actually change the methods now
    return imm
def _imm_trans_to_persist(imm):
    # changes state from transient to persistent
    params = _imm_param_data(imm)
    if _imm_is_init(imm):
        _imm_init_to_trans(imm)
    elif _imm_is_persist(imm):
        raise RuntimeError('Attempted to change persistent immutable from transient to persistent')
    # There are no checks to pass at this point, just gotta change the methods:
    #imm.__dict__['__setattr__'] = _imm_setattr
    #imm.__dict__['__delattr__'] = _imm_delattr
    dd = object.__getattribute__(imm, '__dict__')
    del dd['_pimms_immutable_is_trans']
    return imm

def imm_transient(imm):
    '''
    imm_transient(imm) yields a duplicate of the given immutable imm that is transient.
    '''
    if not is_imm(imm):
        raise ValueError('Non-immutable given to imm_transient')
    # make a duplicate immutable that is in the transient state
    dup = copy.copy(imm)
    if _imm_is_init(imm):
        # this is an initial-state immutable...
        _imm_init_to_trans(dup)
    elif _imm_is_persist(imm):
        # it's persistent; re-transient-ize this new one
        dd = object.__getattribute__(dup, '__dict__')
        dd['_pimms_immutable_is_trans'] = True
    return dup
def imm_persist(imm):
    '''
    imm_persist(imm) turns imm from a transient into a persistent immutable and returns imm. If imm
    is already persistent, then it is simply returned.
    '''
    if not is_imm(imm):
        raise ValueError('imm_persist given non-immutable')
    if not _imm_is_persist(imm):
        _imm_trans_to_persist(imm)
    return imm
def imm_copy(imm, **kwargs):
    '''
    imm_copy(imm, a=b, c=d...) yields a persisent copy of the immutable object imm that differs from
    imm only in that the parameters a, c, etc. have been changed to have the values b, d, etc.
    If the object imm is persistent and no changes are made, imm is returned. If imm is transient,
    a persistent copy of imm is always made.
    '''
    if not is_imm(imm):
        raise ValueError('Non-immutable given to imm_copy')
    if imm_is_persistent(imm) and len(kwargs) == 0:
        # no changes and copy risk
        return imm
    dup = copy.copy(imm)
    if _imm_is_trans(dup): dup = dup.persist()
    dd = object.__getattribute__(dup, '__dict__')
    if _imm_is_persist(dup):
        # we update values directly then recompute checks and invalidate cache
        all_checks = set([])
        all_deps = set([])
        params = _imm_param_data(imm)
        for (p,v) in six.iteritems(kwargs):
            if p not in params:
                raise ValueError('attempt to set non-parameter \'%s\' in imm_copy()' % p)
            (_, tx_fn, arg_lists, check_fns, deps) = params[p]
            for (arg_list, check_fn) in zip(arg_lists, check_fns):
                all_checks.add((tuple(arg_list), check_fn))
            all_deps |= set(deps)
            dd[p] = v if tx_fn is None else tx_fn(v)
        # now invalidate the deps
        for dep in all_deps:
            if dep in dd:
                del dd[dep]
        # now run the tests
        for (arg_list, check_fn) in all_checks:
            if not check_fn(*[getattr(dup, arg) for arg in arg_list]):
                raise ValueError(
                    'Requirement \'%s%s\' failed when copying immutable' % (check_fn, arg_list))
    else:
        # this is an initial-state immutable...
        for (p,v) in six.iteritems(kwargs): object.__setattr__(dup, p, v)
        _imm_init_to_trans(dup)
        _imm_trans_to_persist(dup)
    return dup
def imm_params(imm):
    '''
    imm_params(imm) yields a dictionary of the parameters of the immutable object imm.
    '''
    return {p: getattr(imm, p) for p in six.iterkeys(_imm_param_data(imm))}
def imm_values(imm):
    '''
    imm_values(imm) yields a dictionary of the values of the immutable object imm. Note that this
    forces all of the values to be reified, so only use it if you want to force execution of all
    lazy values.
    '''
    return {p: getattr(imm, p) for p in six.iterkeys(_imm_value_data(imm))}
def imm_dict(imm):
    '''
    imm_dict(imm) yields a persistent dictionary of the params and values of the immutable
    object im. Note that this forces all of the values to be reified, so only use it if you want to
    force execution of all lazy values.
    '''
    immd = dict(**imm_params(imm))
    for (k,v) in six.iteritems(imm_values(imm)):
        immd[k] = v
    return immd
def imm_is_persistent(imm):
    '''
    imm_is_persistent(imm) yields True if imm is a persistent immutable object, otherwise False.
    '''
    return is_imm(imm) and _imm_is_persist(imm)
def imm_is_transient(imm):
    '''
    imm_is_transient(imm) yields True if imm is a transient immutable object, otherwise False.
    '''
    return is_imm(imm) and not _imm_is_persist(imm)

def value(f):
    '''
    The @value decorator, usable in an immutable class (see immutable), specifies that the following
    function is actually a calculator for a lazy value. The function parameters are the attributes
    of the object that are part of the calculation.
    '''
    (args, varargs, kwargs, dflts) = getargspec_py27like(f)
    if varargs is not None or kwargs is not None or dflts:
        raise ValueError('Values may not accept variable, variadic keyword, or default arguments')
    f._pimms_immutable_data_ = {}
    f._pimms_immutable_data_['is_value'] = True
    f._pimms_immutable_data_['inputs'] = args
    f._pimms_immutable_data_['name'] = f.__name__
    f = staticmethod(f)
    return f
def param(f):
    '''
    The @param decorator, usable in an immutable class (see immutable), specifies that the following
    function is actually a transformation on an input parameter; the parameter is required, and is 
    set to the value returned by the function decorated by the parameter; i.e., if you decorate the
    function abc with @param, then imm.abc = x will result in imm's abc attribute being set to the
    value of type(imm).abc(x).
    '''
    (args, varargs, kwargs, dflts) = getargspec_py27like(f)
    if varargs is not None or kwargs is not None or dflts:
        raise ValueError('Params may not accept variable, variadic keyword, or default arguments')
    if len(args) != 1:
        raise ValueError('Parameter transformation functions must take exactly one argument')
    f._pimms_immutable_data_ = {}
    f._pimms_immutable_data_['is_param'] = True
    f._pimms_immutable_data_['name'] = f.__name__
    f = staticmethod(f)
    return f
def option(default_value):
    '''
    The @option(x) decorator, usable in an immutable class (see immutable), is identical to the
    @param decorator except that the parameter is not required and instead takes on the default
    value x when the immutable is created.
    '''
    def _option(f):
        (args, varargs, kwargs, dflts) = getargspec_py27like(f)
        if varargs is not None or kwargs is not None or dflts:
            raise ValueError(
                'Options may not accept variable, variadic keyword, or default arguments')
        if len(args) != 1:
            raise ValueError('Parameter transformation functions must take exactly one argument')
        f._pimms_immutable_data_ = {}
        f._pimms_immutable_data_['is_param'] = True
        f._pimms_immutable_data_['default_value'] = default_value
        f._pimms_immutable_data_['name'] = f.__name__
        f = staticmethod(f)
        return f
    return _option
def require(f):
    '''
    The @require decorator, usable in an immutable class (see immutable), specifies that the
    following function is actually a validation check on the immutable class. These functions
    will appear as static members of the class and get called automatically when the relevant
    data change. Daughter classes can overload requirements to change them, or may add new
    requirements with different function names.
    '''
    (args, varargs, kwargs, dflts) = getargspec_py27like(f)
    if varargs is not None or kwargs is not None or dflts:
        raise ValueError(
            'Requirements may not accept variable, variadic keyword, or default arguments')
    f._pimms_immutable_data_ = {}
    f._pimms_immutable_data_['is_check'] = True
    f._pimms_immutable_data_['inputs'] = args
    f._pimms_immutable_data_['name'] = f.__name__
    f = staticmethod(f)
    return f

def _imm_trans_clos(edges):
    closure = {k: set([]) for e in edges for k in e}
    for (src,dst) in edges:   closure[src].add(dst)
    for (src,dst) in six.iteritems(closure):
        if len(dst) == 0:
            closure[src] = []
    running = True
    while running:
        running = False
        for (src,dsts) in six.iteritems(closure):
            if isinstance(dsts, list): continue
            new_dsts = dsts.copy()
            for dst in dsts:
                further_dsts = closure[dst]
                if isinstance(further_dsts, list):
                    new_dsts |= set(further_dsts)
                else:
                    new_dsts |= further_dsts
                    running = True
            closure[src] = list(dsts) if new_dsts == dsts else new_dsts
    return set([(src,dst) for (src,dsts) in six.iteritems(closure) for dst in dsts])
def _imm_resolve_deps(cls):
    '''
    _imm_resolve_deps(imm_class) resolves the dependencies of the given immutable class imm_class
    and edits the immutable metadata appropriately.
    '''
    dat = cls._pimms_immutable_data_
    params = dat['params']
    values = dat['values']
    consts = dat['consts']
    checks = dat['checks']
    members = list(params.keys()) + list(values.keys())
    mem_ids = {k:i for (i,k) in enumerate(members)}
    # make sure that every input that's not already a value or param becomes a param:
    all_inputs = [v[0] for v in six.itervalues(values)] + [c[0] for c in six.itervalues(checks)]
    all_inputs = set([i for inp in all_inputs for i in inp])
    extra_inputs = [i for i in all_inputs if i not in mem_ids]
    for i in extra_inputs:
        params[i] = (None, None, [], [], [])
        mem_ids[i] = len(members)
        members.append(i)
    # create a graph of the dependencies:
    dep_edges = set([])
    for (v,(inputs,_,_)) in six.iteritems(values):
        for i in inputs:
            dep_edges.add((mem_ids[v], mem_ids[i]))
    # get the transitive closure...
    deps = _imm_trans_clos(dep_edges)
    # we can put all the param and value deps into their appropriate places now
    for (dependant, dependency) in deps:
        if dependency is dependant:
            raise RuntimeError('circular dependency in immutable: value \'%s\'' % dependant)
        (mdpcy, mdpdt) = (members[dependency], members[dependant])
        if mdpcy in params:
            params[mdpcy][4].append(mdpdt)
        elif mdpcy in values:
            values[mdpcy][2].append(mdpdt)
    # last major task is to setup the checks
    deps2params = {v: set([]) for v in six.iterkeys(values)}
    for (p,pd) in six.iteritems(params):
        for v in pd[4]:
            deps2params[v].add(p)
    deps2consts = {v: set([]) for v in six.iterkeys(values)}
    for c in six.iterkeys(consts):
        deps = values[c][2]
        for v in deps:
            deps2consts[v].add(c)
    for (c,(arg_list,check_fn)) in six.iteritems(checks):
        param_list = set([])
        const_list = set([])
        for a in arg_list:
            if a in params: param_list.add(a)
            elif a in values:
                if a in consts: const_list.add(a)
                else:
                    param_list |= deps2params[a]
                    const_list |= deps2consts[a]
            else:
                raise RuntimeError('requirement %s requested non-member: %s' % (c, a))
        for p in param_list:
            params[p][2].append(arg_list)
            params[p][3].append(check_fn)
        for c in const_list:
            consts[p][0].append(arg_list)
            consts[p][1].append(check_fn)
    # That's it; all data should be built at this point
    return cls
def _imm_merge_class(cls, parent):
    '''
    _imm_merge_class(imm_class, parent) updates the given immutable class imm_class to have the
    appropriate attributes of its given parent class. The parents should be passed through this
    function in method-resolution order.
    '''
    # If this is not an immutable parent, ignore it
    if not hasattr(parent, '_pimms_immutable_data_'): return cls
    # otherwise, let's look at the data
    cdat = cls._pimms_immutable_data_
    pdat = parent._pimms_immutable_data_
    # for params, values, and checks, we add them to cls only if they do not already exist in cls
    cparams = cdat['params']
    cvalues = cdat['values']
    cconsts = cdat['consts']
    for (param, (dflt, tx_fn, arg_lists, check_fns, deps)) in six.iteritems(pdat['params']):
        if param not in cparams and param not in cvalues:
            cparams[param] = (dflt, tx_fn, [], [], [])
    for (value, (arg_list, calc_fn, deps)) in six.iteritems(pdat['values']):
        if value in cparams:
            raise ValueError('cannot convert value into parameter: %s' % value)
        if value not in cvalues:
            cvalues[value] = (arg_list, calc_fn, [])
            if len(arg_list) == 0:
                cconsts[value] = ([], [])
    cchecks = cdat['checks']
    for (check, (arg_list, check_fn)) in six.iteritems(pdat['checks']):
        if check not in cchecks:
            cchecks[check] = (arg_list, check_fn)
    # That's it for now
    return cls
def _annotate_imm(cls):
    '''
    _annotate_imm(cls) crawls the class members and adds the _pimms_immutable_data_ attribute
    (a dict) with the following items to the class's attributes:
      * 'params'
      * 'values'
      * 'checks'
    '''
    dat = {}
    cls._pimms_immutable_data_ = dat
    # First, crawl this class's static members to see what has been added...
    mems = inspect.getmembers(cls)
    immmems = [m for m in mems
               if isinstance(m[1], types.FunctionType)
               if hasattr(m[1], '_pimms_immutable_data_')]
    dat['params'] = {}
    dat['values'] = {}
    dat['checks'] = {}
    dat['consts'] = {}
    for (fname,f) in immmems:
        if 'is_param' in f._pimms_immutable_data_:
            dat['params'][f.__name__] = (
                None if 'default_value' not in dat else (dat['default_value'],),
                f, [], [], [])
        elif 'is_value' in f._pimms_immutable_data_:
            inputs = f._pimms_immutable_data_['inputs']
            dat['values'][f.__name__] = (inputs, f, [])
            if len(inputs) == 0:
                dat['consts'][f.__name__] = ([], [])
        elif 'is_check' in f._pimms_immutable_data_:
            dat['checks'][f.__name__] = (f._pimms_immutable_data_['inputs'], f)
    # That grabs all the relevant immutable data; we now have two tasks:
    # (1) merge with parent immutable classes to generate final value/param/check lists
    # --- to do this, we start by getting the class hierarchy
    mro = [c for c in inspect.getmro(cls) if c is not cls and is_imm_type(c)]
    # --- now walk through, merging each
    for parent in mro: _imm_merge_class(cls, parent)
    # (2) resolve dependencies
    _imm_resolve_deps(cls)
    return cls

# depending on python 2/3 we need to make the methods a certain way...
if six.PY2:
    def _method_type(f, cls):
        return types.MethodType(f, None, cls)
else:
    def _method_type(f, cls):
        return f.__get__(None, cls)

def immutable(cls):
    '''
    The @immutable decorator makes an abstract type out of the decorated class that overloads
    __new__ to create interesting behavior consistent with immutable data types. The following
    decorators may be used inside of the decorated class to define immutable behavior:
      * @value indicates that the following function is really a value that should be calculated
        and stored as a value of its arguments. The arguments should not start with self and 
        should instead be named after other values from which it is calculated. If there are no
        arguments, then the returned value is a constant. Note that self is not an argument to this
        function.
      * @param indicates that the following function is really a variable that should be checked
        by the decorated function. Params are settable as long as the immutable object is transient.
        The check function decorated by @param() is actually a transformation function that is
        called every time the parameter gets changed; the actual value given to which the param is
        set is the value returned by this function. The function may raise exceptions to flag
        errors. Note that self is not an argument to this function. All parameters are required for
        an instantiated object; this means that all parameters must either be provided as values or
        options of implementing classes or must be assigned in the constructor.
      * @option(x) indicates that the following function is really an optional value; the syntax and
        behavior of @option is identical to @param except that @option(x) indicates that, if not
        provided, the parameter should take the value x, while @param indicates that an exception
        should be raised.
      * @require indicates that the following function is a requirement that should be run on the
        given arguments (which should name params/options/values of the class). Note that self is
        an argument to the function. If the function yields a truthy value, then the requirement is
        considered to be met; if it raises an exception or yields a non-trithy value (like None or
        []), then the requirement is not met and the object is considered invalid.
    In immutable objects, the functions defined by @require decorators are not instantiated; they
    may, however, be overloaded and called back to the parent class.
    '''
    # Annotate the class!
    cls = _annotate_imm(cls)
    # The attributes we want to make sure get set:
    auto_members = (('__getattribute__', _imm_getattribute),
                    ('__setattr__',      _imm_setattr),
                    ('__delattr__',      _imm_delattr),
                    ('__copy__',         _imm__copy__),
                    ('__deepcopy__',     _imm__copy__))
    for (name, fn) in auto_members: setattr(cls, name, _method_type(fn, cls))
    # __new__ is special...
    @staticmethod
    def _custom_new(c, *args, **kwargs): return _imm_new(c)
    setattr(cls, '__new__', _custom_new)
    # and the attributes we set only if they haven't been specified
    optl_members = (('is_persistent',    _imm_is_persist),
                    ('is_transient',     _imm_is_trans),
                    ('persist',          imm_persist),
                    ('transient',        imm_transient),
                    ('copy',             imm_copy),
                    ('params',           imm_params),
                    ('values',           imm_values),
                    ('todict',           imm_dict))
    for (name, fn) in optl_members:
        if not hasattr(cls, name):
            setattr(cls, name, _method_type(fn, cls))
    # and the attributes we set if they're not overloaded from object
    initfn = _imm_default_init if cls.__init__ is object.__init__ else cls.__init__
    # we set this up so that it can monitor entry and exit from this specific class's __init__()
    def _init_wrapper(imm, *args, **kwargs):
        # call the init normally...
        initfn(imm, *args, **kwargs)
        # If we're still initializing after running the constructor, we need to switch to
        # transient
        if type(imm) is cls and _imm_is_init(imm): _imm_init_to_trans(imm)
        # Okay, all checks passed!
    setattr(cls, '__init__', _method_type(_init_wrapper, cls))
    dflt_members = (('__dir__',          _imm_dir),
                    ('__repr__',         _imm_repr),
                    ('__hash__',         _imm_hash))
    for (name, fn) in dflt_members:
        if not hasattr(cls, name) or not hasattr(object, name) or \
           getattr(cls, name) is getattr(object, name):
            setattr(cls, name, _method_type(fn, cls))
    # Done!
    return cls



####################################################################################################
# pimms/util.py
# Utility classes for functional programming with pimms!
# By Noah C. Benson

import inspect, types, sys, six, pint, os, numbers
import collections as colls, numpy as np, pyrsistent as ps
import scipy.sparse as sps
from functools import (reduce, partial)
if six.PY2:
    try:              from cStringIO import StringIO as BytesIO
    except Exception: from StringIO  import StringIO as BytesIO
else:
    from io import BytesIO
try:              from six.moves import cPickle as pickle
except Exception: import pickle

if six.PY2: tuple_type = types.TupleType
else:       tuple_type = tuple
if six.PY2: list_type = types.ListType
else:       list_type = list

# We want to disable the awful pint warning fo numpy if it's present:
with warnings.catch_warnings():
    warnings.simplefilter("ignore")
    Quantity([])

units = pint.UnitRegistry()
if not hasattr(units, 'pixels'):
    units.define('pixel = [image_length] = px')

if six.PY2:
    def getargspec_py27like(f):
        '''
    getargspec_py27like(f) yields the results of calling inspect.getargspec(f) in Python 2.7, or the
      equivalent in Python 3.
    '''
        return inspect.getargspec(f)
else:
    def getargspec_py27like(f):
        '''
    getargspec_py27like(f) yields the results of calling inspect.getargspec(f) in Python 2.7, or the
      equivalent in Python 3.
    '''
        return inspect.getfullargspec(f)[:4]

def is_unit(q):
    '''
    is_unit(q) yields True if q is a pint unit or a string that names a pint unit and False
      otherwise.
    '''
    if isinstance(q, six.string_types):
        try: return hasattr(units, q)
        except Exception: return False
    else:
        cls = type(q)
        return cls.__module__.startswith('pint.') and cls.__name__ == 'Unit'
def is_quantity(q):
    '''
    is_quantity(q) yields True if q is a pint quantity or a tuple (scalar, unit) and False
      otherwise.
    '''
    if isinstance(q, tuple):
        return len(q) == 2 and is_unit(q[1])
    else:
        cls = type(q)
        return cls.__module__.startswith('pint.') and cls.__name__ == 'Quantity'
def unit(u):
    '''
    unit(u) yields the pimms-library unit object for the given unit object u (which may be from a
      separate pint.UnitRegistry instance).
    unit(uname) yields the unit object for the given unit name uname.
    unit(None) yields None.
    unit(q) yields the unit of the given quantity q.
    '''
    if u is None:    return None
    elif is_unit(u): return getattr(units, str(u))
    elif is_quantity(u):
        if isinstance(u, tuple): return getattr(units, str(u[1]))
        else: return getattr(units, str(u.u))
    else:
        raise ValueError('unrecotnized unit argument')
def mag(val, u=Ellipsis):
    '''
    mag(scalar) yields scalar for the given scalar (numeric value without a unit).
    mag(quant) yields the scalar magnitude of the quantity quant.
    mag(scalar, u=u) yields scalar.
    mag(quant, u=u) yields the given magnitude of the given quant after being translated into the
      given unit u.

    The quant arguments in the above list may be replaced with (scalar, unit) as in (10, 'degrees')
    or (14, 'mm'). Note that mag always translates all lists and tuples into numpy ndarrays.
    '''
    if not is_quantity(val): return val
    if isinstance(val, tuple):
        val = units.Quantity(val[0], val[1])
    return val.m if u is Ellipsis else val.to(unit(u)).m
def imm_array(q):
    '''
    imm_array(val) yields a version of val that is wrapped in numpy's ndarray class. If val is a
      read-only numpy array, then it is returned as-is; if it is not, then it is cast copied to a
      numpy array, duplicated, the read-only flag is set on the new object, and then it is returned.
    '''
    if is_quantity(q):
        m = mag(q)
        mm = imm_array(m)
        return q if mm is m else units.Quantity(mm, unit(q))
    elif not isinstance(q, np.ndarray) or q.flags['WRITEABLE']:
        q = np.array(q)
        q.setflags(write=False)
    return q
def quant(val, u=Ellipsis):
    '''
    quant(scalar) yields a dimensionless quantity with the magnitude given by scalar.
    quant(q) yields q for any quantity q; if q is not part of the pimms units registry, then a
      version of q registered to pimms.units is yielded. Note that q is a quantity if
      pimms.is_quantity(q) is True.
    quant(x, u) yields the given scalar or quantity x converted to the given unit u; if x is a
      scalar or a dimensionless quantity, then the unit u is given to the new quantity with no
      conversion; otherwise, x must be a quantity whose unit can be converted into the unit u.
    '''
    if is_quantity(val):
        if isinstance(val, tuple) or val._REGISTRY is not units:
            val = units.Quantity(mag(val), unit(val))
        return val if u is Ellipsis or u is None else val.to(unit(u))
    else:
        return units.Quantity(val, units.dimensionless if u is Ellipsis or u is None else unit(u))
def iquant(val, u=Ellipsis):
    '''
    iquant(...) is equivalent to quant(...) except that the magnitude of the return value is always
      a read-only numpy array object.
    '''
    if u is not Ellipsis and u is not None: u = unit(u)
    if is_quantity(val):
        uu = unit(val)
        if u is Ellipsis or u == uu:
            # no conversion necessary; might be able to reuse old array
            m  = mag(val)
            mm = imm_array(m)
            if m is not mm or isinstance(val, tuple) or val._REGISTRY is not units:
                val = units.Quantity(mm, uu)
            return val
        else:
            # we convert to another type first, then make an imm array
            if isinstance(val, tuple) or val._REGISTRY is not units:
                val = units.Quantity(mag(val), uu)
            v = val.to(u)
            return units.Quantity(imm_array(v.m), v.u)
    else:
        return units.Quantity(imm_array(val), units.dimensionless if u is Ellipsis else unit(u))
def like_units(a, b):
    '''
    like_units(a,b) yields True if a and b can be cast to each other in terms of units and False
      otherwise. Non-united units are considered dimensionless units.
    '''
    a = quant(0.0, a) if is_unit(a) else a if is_quantity(a) else quant(a, units.dimensionless)
    b = quant(0.0, b) if is_unit(b) else b if is_quantity(b) else quant(b, units.dimensionless)
    if a == b: return True
    try:
        c = a.to(b.u)
        return True
    except Exception:
        return False
def qhashform(o):
    '''
    qhashform(o) yields a version of o, if possible, that yields a hash that can be reproduced
      across instances. This correctly handles quantities and numpy arrays, among other things.
    '''
    if is_quantity(o): return ('__#quant', qhashform(mag(o)), str(o.u))
    elif isinstance(o, np.ndarray) and np.issubdtype(o.dtype, np.dtype(np.number).type):
        return ('__#ndarray', o.tobytes())
    elif isinstance(o, (set, frozenset)): return ('__#set', tuple([qhashform(x) for x in o]))
    elif is_map(o): return ps.pmap({qhashform(k):qhashform(v) for (k,v) in six.iteritems(o)})
    elif is_str(o): return o
    elif hasattr(o, '__iter__'): return tuple([qhashform(u) for u in o])
    else: return o
def qhash(o):
    '''
    qhash(o) is a hash function that operates like hash(o) but attempts to, where possible, hash
      quantities in a useful way. It also correctly handles numpy arrays and various other normally
      mutable and/or unhashable objects.
    '''
    return hash(qhashform(o))
_pickle_load_options = {} if six.PY2 else {'encoding': 'latin1'}
io_formats = colls.OrderedDict(
    [('numpy', {
        'match': lambda o:   (isinstance(o, np.ndarray) and
                              np.issubdtype(o.dtype, np.dtype(np.number).type)),
        'write': lambda s,o: np.save(s, o),
        'read':  lambda s:   np.load(s)}),
     ('pickle', {
        'match': lambda o:   True,
        'write': lambda s,o: pickle.dump(o, s),
        'read':  lambda s:   pickle.load(s, **_pickle_load_options)})])
def _check_io_format(obj, fmt):
    try:              return io_formats[fmt]['match'](obj)
    except Exception: return False
def _save_stream_format(stream, obj, fmt):
    fdat = io_formats[fmt]
    fdat['write'](stream, obj)
def _load_stream_format(stream, fmt):
    fdat = io_formats[fmt]
    return fdat['read'](stream)
def _save_stream(stream, obj):
    for (fmt,fdat) in six.iteritems(io_formats):
        if not _check_io_format(obj, fmt): continue
        s = BytesIO()
        try:
            _save_stream_format(s, obj, fmt)
            pickle.dump(fmt, stream)
            stream.write(s.getvalue())
        except Exception:  continue
        else:              return stream
        finally:           s.close()
    raise ValueError('unsavable object: did not match any exporters')
def _load_stream(stream):
    try:              fmt = pickle.load(stream, **_pickle_load_options)
    except Exception: raise ValueError('could not unpickle format; probably not a pimms save file')
    if not isinstance(fmt, six.string_types):
        raise ValueError('file format object is not a string; probably not a pimms save file')
    if fmt not in io_formats:
        raise ValueError('file has unrecognized format \'%s\'' % fmt)
    return _load_stream_format(stream, fmt)
def save(filename, obj, overwrite=False, create_directories=False, create_mode=0o755):
    '''
    pimms.save(filename, obj) attempts to pickle the given object obj in the filename (or stream,
      if given). An error is raised when this cannot be accomplished; the first argument is always
      returned; though if the argument is a filename, it may be a differet string that refers to
      the same file.

    The save/load protocol uses pickle for all saving/loading except when the object is a numpy
    object, in which case it is written using obj.tofile(). The save function writes meta-data into
    the file so it cannot simply be unpickled, but must be loaded using the pimms.load() function.
    Fundamentally, however, if an object can be pickled, it can be saved/loaded by pimms.

    Options:
      * overwrite (False) The optional parameter overwrite indicates whether an error should be
        raised before opening the file if the file already exists.
      * create_directories (False) The optional parameter create_directories indicates whether the
        function should attempt to create the directories in which the filename exists if they do
        not already exist.
      * create_mode (0o755) The optional parameter create_mode specifies the mode given to
        directories that are created by the save function (if any).
    '''
    if isinstance(filename, six.string_types):
        filename = os.path.expanduser(filename)
        if not overwrite and os.path.exists(filename):
            raise ValueError('save would overwrite file %s' % filename)
        if create_directories:
            dname = os.path.dirname(os.path.realpath(filename))
            if not os.path.isdir(dname): os.makedirs(dname, mode=create_mode)
        with open(filename, 'wb') as f:
            _save_stream(f, obj)
    else:
        _save_stream(filename, obj)
    return filename
def load(filename, ureg='pimms'):
    '''
    pimms.load(filename) loads a pimms-formatted save-file from the given filename, which may
      optionaly be a string. By default, this function forces all quantities (via the pint
      module) to be loaded using the pimms.units unit registry; the option ureg can change
      this.

    If the filename is not a correctly formatted pimms save-file, an error is raised.

    Options:
      * ureg ('pimms') specifies the unit-registry to use for ping module units that are loaded
        from the files; 'pimms' is equivalent to pimms.units. None is equivalent to using the
        pint._APP_REGISTRY unit registry.
    '''
    if ureg is not None:
        if ureg == 'pimms': ureg = units
        orig_app_ureg = pint._APP_REGISTRY
        orig_dfl_ureg = pint._DEFAULT_REGISTRY
        pint._APP_REGISTRY = ureg
        pint._DEFAULT_REGISTRY = ureg
        try:              return load(filename, ureg=None)
        except Exception: raise
        finally:
            pint._APP_REGISTRY     = orig_app_ureg
            pint._DEFAULT_REGISTRY = orig_dfl_ureg
    if isinstance(filename, six.string_types):
        filename = os.path.expanduser(filename)
        with open(filename, 'rb') as f:
            return _load_stream(f)
    else:
        return _load_stream(filename)

def _load_json(filename, to='auto'):
    '''
    _load_json(filename) yields the object represented by the json file or stream object filename.
    
    The optional argument to may be set to None to indicate that the JSON data should be returned
    verbatim rather than parsed by neuropythy's denormalize system.
    '''
    import json
    if is_str(filename):
        try:
            with gzip.open(filename, 'rt') as fl: return json.load(fl)
        except Exception: pass
        with open(filename, 'rt') as fl: return json.load(fl)
    else: return json.load(filename)
def to_jsonable(obj):
    '''
    to_jsonable(obj) attempts to convert the given obj into an object that can be encoded into
      JSON using the json package's dumps() function. Essentially, this converts persistent
      datatypes into the typical python datatypes (e.g., pyrsistent.pmap -> dict,
      pyrsistent.pvec -> list). 
    '''
    if   is_map(obj): return dict({k: to_jsonable(v) for (k,v) in six.iteritems(obj)})
    elif is_seq(obj): return [to_jsonable(u) for u in obj]
    else: return obj
def cache_filename(cache_path, params, suffix='.pp', create_mode=0o755):
    '''
    cache_filename(cache_path, params, ext) yields a path p such that p is a file in the given
      directory cache_path that is associated with the given set of params such that a repeated
      call to find_cache_file with the same arguments will return the same file for as long as
      the cache_path is not cleared. The returned file has the given extention.
    cache_filename(cache_path, params) uses the extention '.pp' (for pimms-pickle).
    '''
    import json
    if not os.path.isdir(cache_path): raise ValueError('cache_path not found: %s' % (cache_path,))
    if suffix is None: suffix = ''
    # for this to work we need to be able to hash h...
    try: h = qhash(params)
    except Exception: h = None
    if h is None: raise ValueError('could not qhash given parameters')
    # ... and we need to be able to convert params into json...
    try:
        nparams = to_jsonable(params)
        j = json.dumps(nparams)
    except Exception: j = None
    if j is None: raise ValueError('could not convert parameters into json')
    # okay, the has determines the cache subdirectory:
    hstr = '%+016x' % (h,)
    hstr = ('p' if hstr[0] == '+' else 'n') + hstr[1:]
    # see if this directory exists...
    cache_path = os.path.join(cache_path, hstr)
    if   not os.path.exists(cache_path): os.makedirs(cache_path, mode=create_mode)
    elif not os.path.isdir(cache_path):  raise ValueError('params cache to a file, not a dir')
    # okay, try to find the file:
    k = 1
    while True:
        code = '%06x' % k
        pfile = os.path.join(cache_path, 'key-' + code + '.json')
        if os.path.isfile(pfile):
            try:
                with open(pfile, 'r') as fl: s = fl.read()
                js = json.loads(s)
                if nparams == js: break
            except Exception: pass
        else:
            # write out this pfile then break
            with open(pfile, 'w') as fl: fl.write(j)
            break
        k += 1
    # okay, return the filename that goes with this code
    return os.path.join(cache_path, code + suffix)

def cache_lmap(cache_path, params, lmap, create_mode=0o755, key_translator=None):
    '''
    cache_lmap(cache_path, params, lmap) yields a lazy-map whose keys and values are identical to
      the given map lmap except that the returned map attempts to load the lazy values in lmap
      from cache prior to calculating them, and saves the lmap values out to cache if they must be
      calculated.

    Note that for any (key,val) pair in lmap, if either val cannot be saved via pimms.save() or if
    key is not a valid filename, then that pair will not be cached. The optional argument
    key_translator may be given (either a map or a function) that converts lmap's keys into
    appropriate strings.

    This function catches and ignores exceptions, so it is safe to wrap a map using cache_lmap, even
    if the params are not hashable or the map's values cannot be saved via pimms.save; any such
    values are simply returned without caching.
    '''
    if not is_map(lmap): raise ValueError('cache_map given non-map object of type %s'%(type(lmap),))
    if not is_lazy_map(lmap): lmap = lazy_map(lmap)
    params = persist(params)
    ktr_key = '__pimms_cache_lmap_key_translation__'
    # if ktr is given, make sure it will work:
    if ktr_key in params: key_translator = params[ktr_key]
    if key_translator is not None:
        if pimms.is_map(key_translator):
            ktr = {k:key_translator.get(k,k) for k in six.iterkeys(lmap)}
        else: ktr = {k:key_translator(k) for k in six.iterkeys(lmap)}
        params = pimms.assoc(params, ktr_key, ps.pmap(ktr))
    else: ktr = ps.m()
    # okay, create the layered lmap:
    def _cache(k0):
        if not lmap.is_lazy(k0): return lmap[k0]
        if _cache.cdir is None:
            try:
                # find a directory for the lmap items:
                _cache.cdir = cache_filename(cache_path, params, '', create_mode=create_mode)
                # make the dir if necessary
                if not os.path.isdir(cdir): os.makedirs(cdir, mode=create_mode)
            except Exception: _cache.cdir = Ellipsis
        k = ktr.get(k0,k0)
        if _cache.cdir is Ellipsis: flnm = None
        else:
            try:
                flnm = os.path.join(_cache.cdir, k + '.pp')
                if os.path.isfile(flnm): return load(flnm)
            except Exception: flnm = None
        v = lmap[k0]
        if flnm is None: return v
        try: save(flnm, v)
        except Exception: pass
        return v
    _cache.cdir = None
    return lazy_map({k: curry(_cache, k) for k in six.iterkeys(lmap)})
def cache_fn(cache_path, params, fn, create_mode=0o755):
    '''
    cache_fn(cache_path, params, fn) yields a function that is equivalent to the given function fn
      except that the returned function first attempts to load the result of calling fn() from
      cache and returns that value if found; if not, fn() is called and the result is cached, if
      possible.

    This function catches and ignores exceptions, so it is always safe to wrap a function using
    cache_fn: even if the caching fails, the returned function will mimic fn.
    '''
    def _fn():
        if _fn._fn is None: return _fn._result
        # first, get the cache filename:
        try:
            # find a directory for the lmap items:
            cfl = cache_filename(cache_path, params, create_mode=create_mode)
            # load it if possible
            if os.path.isfile(cfl): return load(cfl)
        except Exception: cfl = Ellipsis
        x = _fn._fn()
        _fn._result = x
        _fn._fn = None # allow garbage collection
        if cfl is Ellipsis: return x
        try: save(cfl, x)
        except Exception: pass
        return x
    _fn._fn = fn
    _fn._result = None
    return _fn                

def is_str(arg):
    '''
    is_str(x) yields True if x is a string object or a 0-dim numpy array of a string and yields
      False otherwise. 
    '''
    return (isinstance(arg, six.string_types) or
            is_npscalar(arg, 'string') or
            is_npvalue(arg, 'string'))
def is_class(arg):
    '''
    is_class(x) yields True if x is a class object and False otherwise.
    '''
    return isinstance(arg, six.class_types)

# handy type translations:
_numpy_type_names = {'bool':    (np.bool_,),
                     'int':     (np.integer, np.bool_),
                     'integer': (np.integer, np.bool_),
                     'float':   (np.floating, np.integer, np.bool_),
                     'real':    (np.floating, np.integer, np.bool_),
                     'complex': (np.number,),
                     'number':  (np.number,),
                     'string':  ((np.bytes_ if six.PY2 else np.unicode_),),
                     'unicode': (np.unicode_,),
                     'bytes':   (np.bytes_,),
                     'chars':   (np.character,),
                     'object':  (np.object_,),
                     'any':     (np.generic,),}
if not six.PY2: unicode = str
def numpy_type(type_id):
    '''
    numpy_type(type) yields a tuple of valid numpy types that can represent the type specified in
      the given type argument. The type argument may be a numpy type such as numpy.signedinteger, in
      which case a tuple containing only it is returned. Alternately, it may be a string that labels
      a numpy type or a builtin type that should be translated to a numpy type (see below).

    Note that numpy_types() does note intend to reproduce the numpy type hierarchy! If you want to
    perform comparisons on the numpy hierarchy, use numpy's functions. Rather, numpy_type()
    represents types in a mathematical hierarchy, so, for example, numpy_type('real') yields the
    tuple (numpy.floating, numpy.integer, numpy.bool_) because all of these are valid real numbers.

    Valid numpy type strings include:
      * 'bool'
      * 'int' / 'integer' (integers and booleans)
      * 'float' / 'real' (integers, booleans, or floating-point)
      * 'complex' / 'number' (integers, booleans, reals, or complex numbers)
      * 'string' ('unicode' in Python 3, 'bytes' in Python 2)
      * 'unicode'
      * 'bytes'
      * 'chars' (numpy.character)
      * 'object'
      * 'any'
    Valid builtin types include:
      * int (also long in Python 2), like 'int'
      * float, like 'float'
      * complex, like 'complex'
      * str, like 'string'
      * bytes, like 'bytes',
      * unicode, like 'unicode'
      * object, like 'object'
      * None, like 'any'
    '''
    if is_str(type_id):
        return _numpy_type_names[type_id.lower()]
    elif isinstance(type_id, (list, tuple)):
        return tuple([dt for s in type_id for dt in numpy_type(s)])
    elif type_id is None:
        return _numpy_type_names['any']
    elif type_id in six.integer_types:       return _numpy_type_names['int']
    elif type_id is float:                   return _numpy_type_names['float']
    elif type_id is numbers.Real:            return _numpy_type_names['real']
    elif type_id is complex:                 return _numpy_type_names['complex']
    elif type_id is numbers.Complex:         return _numpy_type_names['complex']
    elif type_id is bool:                    return _numpy_type_names['bool']
    elif type_id is str:                     return _numpy_type_names['string']
    elif type_id is unicode:                 return _numpy_type_names['unicode']
    elif type_id is bytes:                   return _numpy_type_names['bytes']
    elif type_id is object:                  return _numpy_type_names['object']
    elif np.issubdtype(type_id, np.generic): return (type_id,)
    else: raise ValueError('Could not deduce numpy type for %s' % type_id)

def is_nparray(u, dtype=None, dims=None):
    '''
    is_nparray(u) yields True if u is an instance of a numpy array and False otherwise.
    is_nparray(u, dtype) yields True if is_nparray(u) and the dtype of u is a sub-dtype of the given
      dtype.
    is_nparray(u, dtype, dims) yields True if is_nparray(u, dtype) and the number of dimensions is
      equal to dims (note that dtype may be set to None for no dtype requirement).
    
    Notes:
      * Either dims or dtype may be None to indicate no requirement; additionally, either may be a
        tuple to indicate that the dtype or dims may be any of the given values.
      * If u is a quantity, then this is equivalent to querying mag(u).

    See also: is_npscalar, is_npvector, is_npmatrix, is_array, is_scalar, is_vector, is_matrix
    '''
    if   is_quantity(u):                return is_nparray(mag(u), dtype=dtype, dims=dims)
    elif not isinstance(u, np.ndarray): return False
    # it's an array... check dtype
    if dtype is not None:
        if not any(np.issubdtype(u.dtype, d) for d in numpy_type(dtype)): return False
    # okay, the dtype is fine; check the dims
    if dims is None: return True
    if isinstance(dims, (tuple,list)): return len(u.shape) in dims
    else:                              return len(u.shape) == dims
def is_npscalar(u, dtype=None):
    '''
    is_npscalar(u) yields True if u is an instance of a numpy array with 0 shape dimensions.
    is_npscalar(u, dtype) additionally requires that u have a dtype that is a sub-dtype of the given
      dtype.

    See also: is_nparray, is_npvector, is_npmatrix, is_array, is_scalar, is_vector, is_matrix
    '''
    return is_nparray(u, dtype=dtype, dims=0)
def is_npvector(u, dtype=None):
    '''
    is_npvector(u) yields True if u is an instance of a numpy array with one shape dimension.
    is_npvector(u, dtype) additionally requires that u have a dtype that is a sub-dtype of the given
      dtype.

    See also: is_nparray, is_npscalar, is_npmatrix, is_array, is_scalar, is_vector, is_matrix
    '''
    return is_nparray(u, dtype=dtype, dims=1)
def is_npmatrix(u, dtype=None):
    '''
    is_npmatrix(u) yields True if u is an instance of a numpy array with two shape dimensions.
    is_npmatrix(u, dtype) additionally requires that u have a dtype that is a sub-dtype of the given
      dtype.

    See also: is_nparray, is_npscalar, is_npvector, is_array, is_scalar, is_vector, is_matrix
    '''
    return is_nparray(u, dtype=dtype, dims=2)
def is_npvalue(u, dtype):
    '''
    is_npvalue(u, dtype) yields True if u is a member of the given dtype according to numpy. The
      dtype may be specified as a string (see numpy_type) or a type. Note that dtype may be None,
      'any', or np.generic, but this will always return True if so.

    Note that is_npvalue(1, 'int') will yield True, while is_npvalue(np.array(1), 'int'),
    is_npvalue(np.array([1]), 'int'), and is_npvalue([1], 'int') will yield False (because lists
    and numpy arrays of ints aren't ints).

    See also is_nparray, is_npscalar, is_scalar.
    '''
    if is_quantity(u): return is_npvalue(mag(u), dtype=dtype)
    return any(np.issubdtype(type(u), np.generic if d is None else d) for d in numpy_type(dtype))

def is_array(u, dtype=None, dims=None):
    '''
    is_array(u) is equivalent to is_nparray(np.asarray(u)), meaning is_array(u) will always yield
      True.
    is_array(u, dtype) yields True if np.asarray(u) is of the given dtype, which is looked up using
      numpy_type. If dtype is None, then no dtype requirement is applied.
    is_array(u, dtype, dims) yields True if np.asarray(u) has the given dtype has the given number
      of dimensions.

    As in is_nparray(), dtype and dims may be tuples to indicate that any of the listed values are
    acceptable.

    See also: is_nparray, is_npscalar, is_npvector, is_npmatrix, is_scalar, is_vector, is_matrix
    '''
    if   is_map(u): return False # sometimes turns into a named-column array
    elif is_quantity(u): return is_array(mag(u), dtype=dtype, dims=dims)
    elif sps.issparse(u): return is_nparray(u[[],[]].toarray(), dtype=dtype, dims=dims)
    else:
        try:              u = np.asarray(u)
        except Exception: pass
        return is_nparray(u, dtype=dtype, dims=dims)
def is_scalar(u, dtype=None):
    '''
    is_scalar(u) is equivalent to is_npscalar(np.asarray(u)).
    is_scalar(u, dtype) is equivalent to is_npscalar(np.asarray(u), dtype).

    See also: is_nparray, is_npscalar, is_npvector, is_npmatrix, is_array, is_vector, is_matrix
    '''
    return is_array(u, dtype=dtype, dims=0)
def is_vector(u, dtype=None):
    '''
    is_vector(u) is equivalent to is_npvector(np.asarray(u)).
    is_vector(u, dtype) is equivalent to is_npvector(np.asarray(u), dtype).

    See also: is_nparray, is_npscalar, is_npvector, is_npmatrix, is_array, is_scalar, is_matrix
    '''
    return is_array(u, dtype=dtype, dims=1)
def is_matrix(u, dtype=None):
    '''
    is_matrix(u) is equivalent to is_npmatrix(np.asarray(u)).
    is_matrix(u, dtype) is equivalent to is_npmatrix(np.asarray(u), dtype).

    See also: is_nparray, is_npscalar, is_npvector, is_npmatrix, is_array, is_scalar, is_vector
    '''
    return is_array(u, dtype=dtype, dims=2)
def is_tuple(arg):
    '''
    is_tuple(arg) yields True if arg is a tuple and False otherwise.
    '''
    return isinstance(arg, tuple_type)
def is_list(arg):
    '''
    is_list(arg) yields True if arg is a list and False otherwise.
    '''
    return isinstance(arg, list_type)
def is_set(arg):
    '''
    is_set(arg) yields True if arg is a set or frozenset and False otherwise.
    '''
    isinstance(arg, colls.Set)
def curry(f, *args0, **kwargs0):
    '''
    curry(f, ...) yields a function equivalent to f with all following arguments and keyword
      arguments passed. This is much like the partial function, but yields a function instead of
      a partial object and thus is suitable for use with pimms lazy maps.
    '''
    def curried_f(*args, **kwargs): 
        return f(*(args0 + args), **merge(kwargs0, kwargs))
    return curried_f
def is_seq(arg):
    '''
    is_seq(arg) yields True if arg is a sequential collection otherwise False; i.e., it must be a
      list, tuple, persistent vector, persistent list, or numpy array.
    
    Note that strings are not considered sequences.
    '''
    return isinstance(arg, (list_type, tuple_type, ps.PVector, ps.PList)) or is_nparray(arg)
def is_pseq(arg):
    '''
    is_pseq(arg) yields True if arg is a persistent vector, a persistent list, a tuple, or a numpy
      array with its writeable flag off.
    '''
    if is_nparray(arg): return arg.flags['WRITEABLE'] == False
    return isinstance(arg, (tuple_type, ps.PVector, ps.PList))
    
def is_int(arg):
    '''
    is_int(x) yields True if x is an integer object and False otherwise; integer objects include the
      standard Python integer types as well as numpy single integer arrays (i.e., where
      x.shape == ()) and quantities with integer magnitudes.
    '''
    return (is_int(mag(arg)) if is_quantity(arg)                   else
            True             if isinstance(arg, six.integer_types) else
            is_npscalar(arg, 'int') or is_npvalue(arg, 'int'))
def is_float(arg):
    '''
    is_float(x) yields True if x is a non-complex numeric object and False otherwise. It is an alias
      for is_real(x).

    Note that is_float(i) will yield True for an integer or bool i; to check for floating-point
    representations of numbers, use is_array(x, numpy.floating) or similar.
    '''
    return (is_float(mag(arg)) if is_quantity(arg)       else
            True               if isinstance(arg, float) else
            is_npscalar(arg, 'real') or is_npvalue(arg, 'real'))
def is_real(arg):
    '''
    is_real(x) yields True if x is a non-complex numeric object and False otherwise.

    Note that is_real(i) will yield True for an integer or bool i; to check for floating-point
    representations of numbers, use is_array(x, numpy.floating) or similar.
    '''
    return (is_real(mag(arg)) if is_quantity(arg)       else
            True              if isinstance(arg, float) else
            is_npscalar(arg, 'real') or is_npvalue(arg, 'real'))
def is_inexact(arg):
    '''
    is_inexact(x) yields True if x is a number represented by floating-point data (i.e., either a
      non-integer real number or a complex number) and False otherwise.
    '''
    return (is_inexact(mag(arg)) if is_quantity(arg) else
            is_npscalar(u, np.inexact) or is_npvalue(arg, np.inexact))
def is_complex(arg):
    '''
    is_complex(x) yields True if x is a complex numeric object and False otherwise. Note that this
      includes anything representable as as a complex number such as an integer or a boolean value.
      In effect, this makes this function an alias for is_number(arg).
    '''
    return (is_complex(mag(arg)) if is_quantity(arg)                 else
            True                 if isinstance(arg, numbers.Complex) else
            is_npscalar(arg, 'complex') or is_npvalue(arg, 'complex'))
def is_number(arg):
    '''
    is_number(x) yields True if x is a numeric object and False otherwise.
    '''
    return (is_number(mag(arg)) if is_quantity(arg)              else
            is_npscalar(arg, 'number') or is_npvalue(arg, 'number'))
def is_map(arg):
    '''
    is_map(x) yields True if x implements Python's builtin Mapping class.
    '''
    return isinstance(arg, colls.Mapping)

class LazyPMap(ps.PMap):
    '''
    LazyPMap is an immutable map that is identical to pyrsistent's PMap, but that treats functions
    of 0 arguments, when values, as lazy values, and memoizes them as it goes.
    '''
    def __init__(self, *args, **kwargs):
        self._memoized = ps.m()
    def __repr__(self):
        s = ', '.join(['%s: %s' % (repr(k), '<lazy>' if self.is_lazy(k) else self[k])
                       for k in self.iterkeys()])
        return 'lmap({' + s + '})'
    def _examine_val(self, k, val):
        'should only be called internally'
        if not isinstance(val, (types.FunctionType, partial)): return val
        vid = id(val)
        if vid in self._memoized:
            return self._memoized[vid]
        elif [] != getargspec_py27like(val)[0]:
            return val
        else:
            val = val()
            object.__setattr__(self, '_memoized', self._memoized.set(vid, val))
            return val
    def __getitem__(self, k):
        return self._examine_val(k, ps.PMap.__getitem__(self, k))
    def iterkeys(self):
        for (k,_) in ps.PMap.iteritems(self):
            yield k
    def iteritems(self):
        for (k,v) in ps.PMap.iteritems(self):
            yield (k, self._examine_val(k, v))
    def iterlazy(self):
        '''
        lmap.iterlazy() yields an iterator over the lazy keys only (memoized lazy keys are not
        considered lazy).
        '''
        for k in self.iterkeys():
            if self.is_lazy(k):
                yield k
    def itermemoized(self):
        '''
        lmap.itermemoized() yields an iterator over the memoized keys only (neihter unmemoized lazy
        keys nor normal keys are considered memoized).
        '''
        for k in self.iterkeys():
            if self.is_memoized(k):
                yield k
    def iternormal(self):
        '''
        lmap.iternormal() yields an iterator over the normal unlazy keys only (memoized lazy keys
        are not considered normal).
        '''
        for k in self.iterkeys():
            if self.is_normal(k):
                yield k
    class _LazyEvolver(ps.PMap._Evolver):
        def persistent(self):
            if self.is_dirty():
                lmap = LazyPMap(self._size, self._buckets_evolver.persistent())
                mems = self._original_pmap._memoized
                for (k,v) in ps.PMap.iteritems(self._original_pmap):
                    if not isinstance(v, (types.FunctionType, partial)): continue
                    vid = id(v)
                    if vid not in mems or ps.PMap.__getitem__(lmap, k) is v: continue
                    mems = mems.discard(vid)
                object.__setattr__(lmap, '_memoized', mems)
                self._original_pmap = lmap
            return self._original_pmap
    def evolver(self):
        return self._LazyEvolver(self)
    def update_with(self, update_fn, *maps):
        evolver = self.evolver()
        for map in maps:
            if isinstance(map, LazyPMap):
                for key in map.iterkeys():
                    value = ps.PMap.__getitem__(map, key)
                    evolver.set(key, update_fn(evolver[key], value) if key in evolver else value)
            else:
                for key, value in map.items():
                    evolver.set(key, update_fn(evolver[key], value) if key in evolver else value)
        return evolver.persistent()
    def is_lazy(self, k):
        '''
        lmap.is_lazy(k) yields True if the given k is lazy and unmemoized in the given lazy map,
        lmap, otherwise False.
        '''
        v = ps.PMap.__getitem__(self, k)
        if not isinstance(v, (types.FunctionType, partial)) or \
           id(v) in self._memoized or \
           [] != getargspec_py27like(v)[0]:
            return False
        else:
            return True
    def is_memoized(self, k):
        '''
        lmap.is_memoized(k) yields True if k is a key in the given lazy map lmap that is both lazy
        and already memoized.
        '''
        v = ps.PMap.__getitem__(self, k)
        if not isinstance(v, (types.FunctionType, partial)):
            return False
        else:
            return id(v) in self._memoized
    def is_normal(self, k):
        '''
        lmap.is_normal(k) yields True if k is a key in the given lazy map lmap that is neither lazy
        nor a formerly-lazy memoized key.
        '''
        v = ps.PMap.__getitem__(self, k)
        if not isinstance(v, (types.FunctionType, partial)) or [] != getargspec_py27like(v)[0]:
            return True
        else:
            return False
colls.Mapping.register(LazyPMap)
colls.Hashable.register(LazyPMap)
def _lazy_turbo_mapping(initial, pre_size):
    '''
    _lazy_turbo_mapping is a blatant copy of the pyrsistent._pmap._turbo_mapping function, except
    it works for lazy maps; this seems like the only way to fully overload PMap.
    '''
    size = pre_size or (2 * len(initial)) or 8
    buckets = size * [None]
    if not isinstance(initial, colls.Mapping): initial = dict(initial)
    for k, v in six.iteritems(initial):
        h = hash(k)
        index = h % size
        bucket = buckets[index]
        if bucket: bucket.append((k, v))
        else:      buckets[index] = [(k, v)]
    return LazyPMap(len(initial), ps.pvector().extend(buckets))
_EMPTY_LMAP = _lazy_turbo_mapping({}, 0)
def lazy_map(initial={}, pre_size=0):
    '''
    lazy_map is a blatant copy of the pyrsistent.pmap function, and is used to create lazy maps.
    '''
    if is_lazy_map(initial): return initial
    if not initial: return _EMPTY_LMAP
    return _lazy_turbo_mapping(initial, pre_size)
def lmap(initial={}, pre_size=0):
    '''
    lmap is an alias for lazy_map.
    '''
    return lazy_map(initial, pre_size)
def is_lazy_map(m):
    '''
    is_lazy_map(m) yields True if m is an instance if LazyPMap and False otherwise. Note that this
      will yield True for a pimms itable object as well as a pimms lazy map because itables respect
      laziness. To check if an object is a pimms lazy map specifically, use
      isinstance(m, pimms.LazyPMap).
    '''
    from .table import is_itable
    return isinstance(m, LazyPMap) or is_itable(m)
def is_lmap(m):
    '''
    is_lmap(m) is identical to is_lazy_map(m).
    '''
    return is_lazy_map(m)
def is_pmap(arg):
    '''
    is_pmap(x) yields True if x is a persistent map object and False otherwise. Note that this will
      yield True for any of the following types: a pyrsistent.PMap (persistent mapping object), a
      pimms lazy map, a pimms itable object. If you want to check specifically if an object is a
      pyrsistent PMap object, use isinstance(arg, pyrsistent.PMap).
    '''
    return isinstance(arg, ps.PMap) or is_lazy_map(arg)

def lazy_value_map(f, m, *args, **kwargs):
    '''
    lazy_value_map(f, mapping) yields a lazy map whose keys are the same as those of the given dict
      or mapping object and whose values, for each key k are f(mapping[k]).
    lazy_value_map(f, mapping, *args, **kw) additionally passes the given arguments to the function
      f, so in the resulting map, each key k is mapped to f(mapping[k], *args, **kw).

    If a dict object that is not persistent is passed to lazy_value_map, then a persistent copy of
    it is made for use with the lazy-map; accordingly, it's not necessary to worry about the 
    persistence of the map you pass to this function. It is, however, important to worry about the
    persistence of the values in the map you pass. If these values are mutated in-place, then the
    lazy map returned from this function could change as well.

    If the given mapping object is an ITable object, then an ITable object is returned.
    '''
    if not is_map(m): raise ValueError('Non-mapping object passed to lazy_value_map')
    if not is_lazy_map(m) and not is_pmap(m): m = ps.pmap(m)
    def curry_fn(k): return lambda:f(m[k], *args, **kwargs)
    m0 = {k:curry_fn(k) for k in six.iterkeys(m)}
    from .table import (is_itable, itable)
    return itable(m0) if is_itable(m) else lazy_map(m0)
def value_map(f, m, *args, **kwargs):
    '''
    value_map(f, mapping) yields a persistent map whose keys are the same as those of the given dict
      or mapping object and whose values, for each key k are f(mapping[k]).
    value_map(f, mapping, *args, **kw) additionally passes the given arguments to the function
      f, so in the resulting map, each key k is mapped to f(mapping[k], *args, **kw).

    Unlike lazy_value_map, this function yields either a persistent or a lazy map depending on the
    input argument mapping. If mapping is a lazy map, then a lazy map is returned; otherwise, a
    persistent non-lazy map is returned.
    '''
    if is_lazy_map(m): return lazy_value_map(f, m, *args, **kwargs)
    else:              return ps.pmap({k:f(v, *args, **kwargs) for (k,v) in six.iteritems(m)})
def key_map(f, m, *args, **kwargs):
    '''
    key_map(f, m) is equivalent to {f(k):v for (k,v) in m.items()} except that it returns a
      persistent mapping object instead of a dict. Additionally, it respects the laziness of maps
      and does not evaluate the values of a lazy map that has been passed to it.
    key_map(f, m, *args, **kwargs) uses f(k, *args, **kwargs) instead of f(k).
    '''
    if is_lazy_map(m):
        from .table import (is_itable, itable)
        def _curry_getval(k): return lambda:m[k]
        m0 = {f(k, *args, **kwargs):_curry_getval(k) for k in six.iterkeys(m)}
        return itable(m0) if is_itable(m) else lazy_map(m0)
    else:
        return ps.pmap({f(k, *args, **kwargs):v for (k,v) in six.iteritems(m)})
def flatten_maps(*args, **kwargs):
    '''
    flatten_maps(*args, **kwags) yields a tuple of the maps in the given arguments; this flattens
      over lists and iterables so long as all elements eventually yield True to is_map(el). The
      optional keyword arguments passed make up the final map.

    This funtion does not evaluate any values of any of the maps and thus implicitly respects the
    laziness of the provided maps.
    '''
    def _recur(arg, work):
        if is_map(arg): work.append(arg)
        elif is_str(arg) or not hasattr(arg, '__iter__'):
            raise ValueError('Non-map given to flatten_maps')
        else:
            for a in arg: _recur(a, work)
    res = []
    for arg in args: _recur(arg, res)
    if len(kwargs) > 0: res.append(kwargs)
    return tuple(res)
def collect(*args, **kwargs):
    '''
    collect(m1, m2, ...) yields a persistent map whose keys are the union of all keys in the given
      maps m1, m2, etc. and whose values are tuples containing each of the given maps (in provided
      order) that contain the given key. This function never evaluates the values in the maps so it
      implicitly supports laziness.

    The collect function fist passes its arguments to flatten_maps, so it is fine to pass lists or
    nested lists of maps to this function; all will be collected.
    '''
    args = flatten_maps(args, **kwargs)
    if len(args) == 0: return ps.m()
    m = {}
    for arg in args:
        for k in six.iterkeys(arg):
            if k in m: m[k].append(arg)
            else: m[k] = [arg]
    return ps.pmap({k:tuple(v) for (k,v) in six.iteritems(m)})
def _choose_first(k, vs):
    '_choose_first(k, vs) yields vs[0][k].'
    return vs[0][k]
def _choose_last(k, vs):
    '_choose_last(k, vs) yields vs[-1][k].'
    return vs[-1][k]
def merge(*args, **kwargs):
    '''
    merge(...) lazily collapses all arguments, which must be python Mapping objects of some kind,
      into a single mapping from left-to-right. The mapping that is returned is a lazy persistent
      object that does not request the value of a key from any of the maps provided until they are
      requested of it; in this fashion it preserves the laziness of immutable map objects that are
      passed to it. Arguments may be mappings or lists/tuples of mappings.

    If all of the arguments passed to merge are pimms itables with the same row_count, then an
    itable object is returned instead of a lazy map.

    The following options are accepted:
    * choose (default None) specifies a function that chooses from which map, of those maps given
      to merge, the value should be drawn when keys overlap. The function is always passed two
      arguments: the key for which the conflict occurs and a list of maps containing that key; it
      should return the value to which the key should be mapped. The default uses the first map.
    '''
    from .table import (is_itable, ITable)
    # figure out the choose-fn
    choose_fn = None
    if 'choose' in kwargs:
        choose_fn = kwargs['choose']
    if len(kwargs) > 1 or (len(kwargs) > 0 and 'choose' not in kwargs):
        raise ValueError('Unidentified options given to merge: %s' (kwargs.keys(),))
    # collect the maps...
    maps = flatten_maps(*args)
    if len(maps) == 0: return ps.m()
    elif len(maps) == 1: return maps[0]
    coll = collect(maps)
    if choose_fn is None: choose_fn = _choose_last
    def curry_choice(k, args): return lambda:choose_fn(k, args)
    resmap = lazy_map({k:curry_choice(k, v) for (k,v) in six.iteritems(coll)})
    # if they're all itables of the same size, return an itable
    if is_itable(maps[0]):
        n = maps[0].row_count
        if all(is_itable(m) and m.row_count == n for m in maps):
            return ITable(resmap, n)
    # otherwise return the lazy map
    return resmap
def lmerge(*args, **kwargs):
    '''
    lmerge(...) is equivalent to merge(...) except for two things: (1) any keyword arguments passed
      to lmerge are bundled into a map that is appended to the argument list to merge, and (2) the
      choose function for the merge always takes the left-most argument. Essentially, this means
      that the map returned from lmerge always chooses the first value it finds in the argument list
      for a particular key.

    See also merge, rmerge.
    '''
    return merge(*(args + (kwargs,)), choose=_choose_first)
def rmerge(*args, **kwargs):
    '''
    rmerge(...) is equivalent to merge(...) except for two things: (1) any keyword arguments passed
      to lmerge are bundled into a map that is appended to the argument list to merge, and (2) the
      choose function for the merge always takes the right-most argument. Essentially, this means
      that the map returned from rmerge always chooses the last value it finds in the argument list
      for a particular key.

    Note that rmerge is the 'traditional' form of the merge operation as found in languages like
    clojure; merge(a,b) overwrites the values of a with any found in b, and merge(a, k=v) is
    roughly equivalent to assoc(a, k=v) (though merge and assoc respect the types of their maps
    slightly differently).

    See also merge, lmerge.
    '''
    return merge(*(args + (kwargs,)), choose=_choose_last)
def is_persistent(arg):
    '''
    is_persistent(x) yields True if x is a persistent object and False if not.

    Note that this persistence can only be checked by the pimms library, so immutable/persistent
    structures not known to pimms or defined in terms of pimms's immutables library cannot be
    evaluated correctly.

    Additionally, is_persistent(x) checks only x; if x is a tuple of mutable objects, it is still
    considered a persistent object.
    '''
    from .immutable import (is_imm, imm_is_persistent)
    if is_imm(arg): return imm_is_persistent(arg)
    elif isinstance(arg, (np.generic, np.ndarray)): return not arg.flags.writeable
    elif is_quantity(arg) and isinstance(mag(arg), (np.generic, np.ndarray)):
        return not mag(arg).flags.writable
    elif is_str(arg): return True
    elif is_number(arg): return True
    elif is_pmap(arg): return True
    elif isinstance(arg, frozenset): return True
    elif isinstance(arg, (ps.PVector, ps.PSet, ps.PList, ps.PRecord)): return True
    else: return False
def persist(arg, depth=Ellipsis, on_mutable=None):
    '''
    persist(x) yields a persistent version of x if possible, or yields x itself.

    The transformations performed by persist(x) are as follows:
      * If x is an immutable object, yields x.persist()
      * If x is a set, yield a frozenset of of persist(u) for all u in x.
      * If x is a numpy array, yield imm_array(x).
      * If x is a map, yields a persistent version of x with all keys and values replaced with their
        persist()'ed form; note that this respects laziness and itables.
      * If x is a list/tuple type, yields a tuple of persist()'ed contents.
      * Otherwise, if the type of x is not recognized, yields x.

    The depth to which persist() searches the argument's elements is controlled by the depth option;
    the default behavior is to persist objects down to the point that a persistent object is found,
    at which its elements are not checked for persistence.

    Note that persist() is not guaranteed to recognize a particular object; it is intended as a
    utility function for basic functional-style and immutable data code in Python. In particular,
    it is usefl for pimms's immutable objects; dicts and mappings; pimms's lazy-maps and itables;
    pyrsistent's sets, vectors, and maps; sets and frozensets (which are both communted to
    frozensets); and anything implementing __iter__ (which are commuted to tuples). Objects that are
    not numbers or strings are considered potentially-mutable and will trigger the on_mutable case.

    The optional arguments may be passed to persist:
      * depth (default: Ellipsis) specifies the depth to which toe persist() function should search
        when persisting objects. The given argument is considered depth 0, so persist(arg, 0) will
        persist only arg and not its contents, if it is a collection. If None is given, then goes to
        any depth; if Ellipsis is given, then searches until a persistent object is found, but does
        not attempt to persist the elements of already-persistent containers (this is the default).
      * on_mutable (default: None) specifies what to do when a non-persistable object is encountered
        in the search. If None, then the object is left; if 'error', then an error is raised;
        otherwise, this must be a function that is passed the object--the return value of this
        function is the replacement used in the object returned from persist().
    '''
    from .immutable import (is_imm, imm_copy)
    # Parse the on_mutable argument
    if on_mutable is None: on_mutable = lambda x:x
    elif on_mutable == 'error':
        def _raise(x):
            raise ValueError('non-persistable: %s' % x)
        on_mutable = _raise
    if depth in (None, Ellipsis): depth_next = depth
    elif depth < 0: return arg
    else: depth_next = depth - 1
    precur = lambda x:persist(x, depth=depth_next, on_mutable=on_mutable)
    # See if we have an easy type to handle
    if is_imm(arg): return imm_copy(arg)
    if is_quantity(arg):
        (m,u) = (mag(arg), unit(arg))
        mm = precur(m)
        if mm is m: return arg
        else: return quant(mm, u)
    elif isinstance(arg, np.ndarray): return imm_array(arg)
    elif isinstance(arg, np.generic):
        x = type(arg)(arg)
        x.setflags(write=False)
        return x
    elif is_str(arg) or is_number(arg): return arg
    elif isinstance(arg, ps.PVector):
        if depth is Ellipsis or depth == 0: return arg
        for (k,v0) in zip(range(len(arg)), arg):
            v = precur(v0)
            if v0 is not v: arg = arg.set(k,v)
        return arg
    elif isinstance(arg, ps.PSet):
        if depth is Ellipsis or depth == 0: return arg
        for v0 in arg:
            v = precur(v0)
            if v0 is not v: arg = arg.discard(v0).add(v)
        return arg
    elif is_pmap(arg):
        if depth is Ellipsis or depth == 0: return arg
        return key_map(precur, value_map(precur, arg))
    elif is_map(arg):
        if not is_pmap(arg): arg = ps.pmap(arg)
        if depth == 0: return arg
        return key_map(precur, value_map(precur, arg))
    elif isinstance(arg, frozenset):
        if depth is Ellipsis or depth == 0: return frozenset(arg)
        a = [x for x in arg]
        q = [precur(x) for x in a]
        if all(ai is qi for (ai,qi) in zip(a,q)): return arg
        return frozenset(q)
    elif isinstance(arg, set):
        if depth == 0: return frozenset(arg)
        a = [x for x in arg]
        q = [precur(x) for x in a]
        if isinstance(arg, frozenset) and all(ai is qi for (ai,qi) in zip(a,q)): return arg
        return frozenset(q)
    elif hasattr(arg, '__iter__'):
        if depth == 0 or (depth is Ellipsis and isinstance(arg, tuple)): return tuple(arg)
        q = tuple(precur(x) for x in arg)
        if isinstance(arg, tuple) and all(ai is qi for (ai,qi) in zip(arg,q)): return arg
        else: return q
    elif isinstance(arg, (types.FunctionType, partial)):
        return arg
    else: return on_mutable(arg)
def dissoc(m, *args, **kw):
    '''
    dissoc(m, k1, k2...) yields a map equivalent to m without the arguments k1, k2, etc. If m is a
      mutable pythin dictionary, this is equivalent to using del m[k] for k in [k1, k2...] then
      returning m itself. If m is a persistent map, this is equivalent to calling m = m.discard(k)
      for all the keys then returning m.

    If m is not a map but is instead a list or persistent vector, this operates as if the
    list/vector is a map whose keys are integers. If an item not at the end of the vector is deleted
    then it is replaced with None instead of pop'ed.

    The optional argument error (default: False) specifies whether an error should be raised when
    one of the keys is not found; this applies to mutable or immutable arguments.
    '''
    if   len(kw) == 0: err = False
    elif len(kw) == 1: err = kw['error']
    else:              raise ValueError('unrecognized optional arguments')
    if is_pmap(m):
        if err:
            for k in args: m = m.remove(k)
        else:
            for k in args: m = m.discard(k)
        return m
    elif is_map(m):
        if err:
            for k in args: del m[k]
        else:
            for k in args:
                if k in m: del m[k]
        return m
    elif isinstance(m, tuple_type):
        return m if len(args) == 0 else tuple(dissoc(list(m), *args, error=err))
    elif isinstance(m, list_type):
        if len(args) == 0: return m
        n = len(m)
        if err:
            for k in args:
                if not is_int(k) or k < 0 or k >= n: raise KeyError(k)
        else: args = [k for k in args if is_int(k)]
        args = sorted(list(set(k if k >= 0 else n + k for k in args)))
        if err:
            for k in args:
                if k < 0 or k >= n: raise KeyError(k)
        else: args = [k for k in args if k < n and k >= 0]
        n -= 1 # just for eliminating the == n-1 in the next line
        while len(args) > 0 and args[-1] == n:
            m.pop()
            args.pop()
            n -= 1
        for k in args: m[k] = None
        return m
    elif is_nparray(m):
        args = list(args)
        if len(args) == 0: return m
        return np.asarray(dissoc(m.tolist(), *args, error=err))
    else: raise ValueError('Cannot dissoc from given type: %s' % type(m))
def assoc(m, *args, **kw):
    '''
    assoc(m, k1, v1, k2, v2...) yields a map equivalent to m without the arguments k1, k2, etc.
      associated with the values v1, v2, etc. If m is a mutable python dictionary, this is
      equivalent to using m[k] = v for all the given key-value pairs then returning m itself. If m
      is a persistent map, this is equivalent to calling m = m.set(k, v) for all the key-value pairs
      then returning m.

    Keys given in the arguments list are always associated in order, then keys in the keyword
    arguments are associated, meaning that the keyword arguments have precedence.

    Note that in addition to key-value pairs in the ordinary arguments list, assoc() also uses the
    key-value pairs given as named/keyword arguments.
    
    If m is not a map but is instead a list or persistent vector, this operates as if the
    list/vector is a map whose keys are integers. If an item is added beyond the extent of the list
    then it is extended by adding None to the list. If negative indices are given, they are always
    interpreted relative to the end of the initial input list m and not in terms of the list as it
    has grown at the point that the key-value pair is processed.
    '''
    if len(args) % 2 != 0: raise ValueError('assoc arguments must be given as key-value pairs')
    args = (u for it in [zip(args[::2],args[1::2]), six.iteritems(kw)] for u in it)
    if is_pmap(m):
        for (k,v) in args: m = m.set(k, v)
        return m
    elif is_map(m):
        for (k,v) in args: m[k] = v
        return m
    elif isinstance(m, tuple_type):
        args = list(args)
        return m if len(args) == 0 else tuple(assoc(list(m), *args, **kw))
    elif isinstance(m, list_type):
        n0 = len(m)
        n = n0
        for (k,v) in args:
            if not is_int(k): TypeError('Keys for list args must be integers')
            if k < -n: KeyError(k)
            if k < 0: k = n0 + k
            if k >= n: (n,m) = (k+1, m + [None]*(k - n + 1))
            m[k] = v
        return m
    elif is_nparray(m):
        args = list(args)
        if len(args) == 0: return m
        return np.asarray(assoc(m.tolist(), *[x for u in args for x in u]))
    else: raise ValueError('Cannot assoc given type: %s' % type(m))
    
                    
            



####################################################################################################
# pimms/cmdline.py
# Immutable Class and functions for parsing command-lines and in particular for parsing a set of
# command-line arguments for use with a pimms calc-plan.
# By Noah C. Benson

import copy, types, os, sys, re, warnings, textwrap, six
import pyrsistent   as     pyr
import numpy        as     np
import collections  as     colls
from   functools    import reduce
from   .immutable   import (immutable, param, value)
from   .calculation import (Plan, is_plan)
from   .util        import (is_map, is_pmap, is_str, is_vector, merge)

@immutable
class CommandLineParser(object):
    '''
    CommandLineParser(instructions) yields a command line parser object, which acts as a function,
      that is capable of parsing command line options into a dictionary of opts and a list of args,
      as defined by the list of instructions.

    The instructions given to a command-line parser should be a list of lists (or tuples), 
    each of which should have three or four entries: [character, word, entry, default] where the
    default is optional. The character and word are the -x and --xxx versions of the argument; the
    entry is the name of the entry in the opts dictionary that is yielded for that command line
    option, and the default is the value inserted into the dictionary if the command line option is
    not found; if no default value is given, then the entry will appear in the dictionary only if it
    appears in the command line arguments.

    If the default value given is either True or False, then the option is understood to be a flag;
    i.e., the option does not take an argument (and single letter flags can appear together, such
    as -it instead of -i -t), and the appearance of the flag toggles the default to the opposite
    value.

    Unless the optional argument value_parse=False is provided, all values are interpreted via the
    ast.literal_eval() function; if this fails or raises an exception then the value is left as a
    string.

    The following optional arguments may be provided to the constructor:
      * value_parse (default: True) specifies whether the values are interpreted via the
        ast.literal_eval() function. This may be set to False to leave the values as strings or it
        may be set to a function that takes one argument and performs the parsing itself; such a
        function f must obey the syntax `parsed_val = f(string_val)`. The value_parse function is
        only called on arguments that have string values included. Note that by default the
        value_parse function interprets the string '...' as Ellipsis in addition to the typical
        ast.literal_eval() behavior.
      * filters (default: None) optionally specifies a dictionary of filter functions, each of which
        is passed the parsed value of the associated argument. Each filter function f must obey the
        syntax `final_value = f(parsed_value)`. The keys of this dictionary must be the entry names
        of the arguments.
    
    Example:
      parser = CommandLineParser(
        [('a', 'flag-a', 'aval', False),
         ('b', 'flag-b', 'bval', True),
         ('c', 'flag-c', 'cval', True),
         ('d', 'flag-d', 'dval', False),
         ('e', 'opt-e',  'eval', None),
         ('f', 'opt-f',  'fval', None),
         ('g', 'opt-g',  'gval'),
         ('h', 'opt-h',  'hval')])
      cmd_line = ['-ab', 'arg1', '--flag-d', '-etestE', '--opt-f=123', '-htestH', 'arg2']
      parser(cmd_line) == parser(*cmd_line)  # identical calls
      parser(cmd_line)
      # ==> (['arg1', 'arg2'],
      # ==>  {'a':True, 'b':False, 'c':True, 'd':True, 'e':'testE', 'f':'123', 'h':'testH'})
    '''

    @staticmethod
    def parse_literal(s):
        '''
        CommandLineParser.parse_literal(s) is equivalent to ast.literal_eval(s) except that (1) it
          additionally interprets '...' to be Ellipsis and (2) when s cannot be parsed as a literal
          it instead returns the string s.
        '''
        from ast import literal_eval
        try:              return literal_eval(s)
        except Exception: return Ellipsis if s.strip() == '...' else s
    
    def __init__(self, instructions, value_parser=True, filters=None):
        'See help(CommandLineParser).'
        wflags = {}
        cflags = {}
        wargs = {}
        cargs = {}
        defaults = {}
        for row in instructions:
            if not hasattr(row, '__iter__') or len(row) < 3 or len(row) > 4 or \
               any(x is not None and not is_str(x) for x in row[:3]):
                raise ValueError('Invalid instruction row: %s ' % row)
            (c, w, var, dflt) = row if len(row) == 4 else (list(row) + [None])
            defaults[var] = dflt
            if dflt is True or dflt is False:
                if c is not None: cflags[c] = var
                if w is not None: wflags[w] = var
            else:
                if c is not None: cargs[c] = var
                if w is not None: wargs[w] = var
        self.default_values = pyr.pmap(defaults)
        self.flag_words = pyr.pmap(wflags)
        self.flag_characters = pyr.pmap(cflags)
        self.option_words = pyr.pmap(wargs)
        self.option_characters = pyr.pmap(cargs)
        self.filters = filters
        self.value_parser = value_parser
    @param
    def filters(fs):
        '''
        parser.filters is a dictionary of filter functions for the various entries in the given
        parser's command line arguments. See also help(CommandLineParser).
        '''
        if not fs: return pyr.m()
        if not is_map(fs): raise ValueError('filters must be a mapping of entries to functions.')
        return pyr.pmap(fs)
    @param
    def value_parser(vp):
        '''
        parser.value_parser is a parse function that accepts a string and yields an interpretation
        of that string as a Python object (which may be the same string or a new value). If this
        value is set to True of Ellipsis, then the default ast.literal_eval()-like behavior is used.
        See also help(CommandLineParser).
        '''
        if vp in [Ellipsis, True]: return CommandLineParser.parse_literal
        elif not vp: return lambda x:x
        else: return vp
    @param
    def default_values(dv):
        '''
        clp.default_values yields the persistent map of default values for the given command-line
          parser clp.
        '''
        if is_pmap(dv): return dv
        elif is_map(dv): return pyr.pmap(dv)
        else: raise ValueError('default_value must be a mapping')
    @param
    def flag_words(u):
        '''
        clp.flag_words yields the persistent map of optional flag words recognized by the given
          command-line parser clp.
        '''
        if is_pmap(u): return u
        elif is_map(u): return pyr.pmap(u)
        else: raise ValueError('flag_words must be a mapping')
    @param
    def flag_characters(u):
        '''
        clp.flag_characters yields the persistent map of the flag characters recognized by the given
          command-line parser clp.
        '''
        if is_pmap(u): return u
        elif is_map(u): return pyr.pmap(u)
        else: raise ValueError('flag_characters must be a mapping')
    @param
    def option_words(u):
        '''
        clp.option_words yields the persistent map of optional words recognized by the given
          command-line parser clp.
        '''
        if is_pmap(u): return u
        elif is_map(u): return pyr.pmap(u)
        else: raise ValueError('option_words must be a mapping')
    @param
    def option_characters(u):
        '''
        clp.option_characters yields the persistent map of optional characters recognized by the
          given command-line parser clp.
        '''
        if is_pmap(u): return u
        elif is_map(u): return pyr.pmap(u)
        else: raise ValueError('option_characters must be a mapping')
        
    def __call__(self, *args):
        if len(args) > 0 and not is_str(args[0]) and is_vector(args[0]):
            args = list(args)
            return self.__call__(*(list(args[0]) + args[1:]))
        parse_state = None
        more_opts = True
        remaining_args = []
        opts = dict(self.default_values)
        wflags = self.flag_words
        cflags = self.flag_characters
        wargs  = self.option_words
        cargs  = self.option_characters
        dflts  = self.default_values
        for arg in [aa for arg in args for aa in (arg if is_vector(arg) else [arg])]:
            larg = arg.lower()
            if parse_state is not None:
                opts[parse_state] = arg
                parse_state = None
            else:
                if arg == '': pass
                elif more_opts and arg[0] == '-':
                    if len(arg) == 1:
                        remaining_args.append(arg)
                    elif arg[1] == '-':
                        trimmed = arg[2:]
                        if trimmed == '':     more_opts = False
                        if trimmed in wflags: opts[wflags[trimmed]] = not dflts[wflags[trimmed]]
                        else:
                            parts = trimmed.split('=')
                            if len(parts) == 1:
                                if trimmed not in wargs:
                                    #raise ValueError('Unrecognized flag/option: %s' % trimmed)
                                    remaining_args.append(arg)
                                else:
                                    # the next argument specifies this one
                                    parse_state = wargs[trimmed]
                            else:
                                k = parts[0]
                                if k not in wargs:
                                    #raise ValueError('Unrecognized option: %s' % k)
                                    remaining_args.append(arg)
                                else:
                                    opts[wargs[k]] = trimmed[(len(k) + 1):]
                    else:
                        trimmed = arg[1:]
                        for (k,c) in enumerate(trimmed):
                            if c in cflags: opts[cflags[c]] = not dflts[cflags[c]]
                            elif c in cargs:
                                remainder = trimmed[(k+1):]
                                if len(remainder) > 0: opts[cargs[c]] = remainder
                                else:
                                    # next argument...
                                    parse_state = cargs[c]
                                break
                else:
                    remaining_args.append(arg)
        if parse_state is not None:
            raise ValueError('Ran out of arguments while awaiting value for %s' % parse_state)
        # that's done; all args are parsed; now we can do the value parsing and the filter functions
        for (k,v) in six.iteritems(opts):
            u = v
            # parse the string, if needed:
            if is_str(v): u = self.value_parser(v)
            # filter if needed:
            if k in self.filters: u = self.filters[k](u)
            # update in options if need-be
            if v is not u: opts[k] = u
        return (remaining_args, opts)

def to_argv_schema(data, arg_names=None, arg_abbrevs=None, filters=None, defaults=None):
    '''
    to_argv_schema(instructions) yields a valid tuple of CommandLineParser instructions for the
      given instructions tuple; by itself, this will only return the instructions as they are, but
      optional arguments (below) will override the values in the instructions if provided.
    to_argv_schema(plan) yields a valid tuple of CommandLineParser instructions for the given plan
      object.

    These schema returned by this function will parse a command-line list (sys.argv) for parameters
    either listed in the instructions (see help(CommandLineParser)) or the afferent parameters of
    the plan. Generally, this should be called by the argv_parse() function and not directly.

    If a plan is given as the first argument, then the following rules are used to determine how
    arguments are parsed:
      * An argument that begins with -- (e.g., --quux-factor=10) is checked for a matching
        plan parameter; the argument name "quux-factor" will match either a parameter called
        "quux-factor" or "quux_factor" (dashes in command-line arguments are auto-translated
        into underscores).
      * If "quux_factor" is a parameter to the plan and is the only parameter that starts with a
        'q', then -q10 or -q 10 are both equivalent to --quux-factor=10. If other parameters
        also start with a 'q' then neither "quux_factor" or the other parameter(s) will be
        matched with the -q flag unless it is specified explicitly via the arg_abbrevs option.
      * Argument values are parsed using Python's ast.literal_eval(); if this raises an
        exception then the value is left as a string.
      * If an argument or flag is provided without an argument (e.g. "--quuxztize" or "-q") then
        it is interpreted as a boolean flag and is given the value True.
      * Arguments that come after the flag "--" are never processed.

    The following options may be given:
      * arg_names (default: None) may be a dictionary that specifies explicity command-line
        argument names for the plan parameters; plan parameters should be keys and the argument
        names should be values. Any parameter not listed in this option will be interpreted
        according to the above rules. If a parameter is mapped to None then it will not be
        filled from the command-line arguments.
      * arg_abbrevs (default:None) may be a dictionary that is handled identically to that of
        arg_names except that its values must be single letters, which are used for the
        abbreviated flag names.
      * defaults (default: None) may specify the default values for the plan parameters; this
        dictionary overrides the default values of the plan itself.
    '''
    if is_plan(data):
        # First we must convert it to a valid instruction list
        (plan, data) = (data, {})
        # we go through the afferent parameters...
        for aff in plan.afferents:
            # these are provided by the parsing mechanism and shouldn't be processed
            if aff in ['argv', 'argv_parsed', 'stdout', 'stderr', 'stdin']: continue
            # we ignore defaults for now
            data[aff] = (None, aff.replace('_', '-'), aff)
        # and let's try to guess at abbreviation names
        entries = sorted(data.keys())
        n = len(entries)
        for (ii,entry) in enumerate(entries):
            if ii > 0   and entry[0] == entries[ii-1][0]: continue
            if ii < n-1 and entry[0] == entries[ii+1][0]: continue
            r = data[entry]
            data[entry] = (entry[0], r[1], entry, r[3]) if len(r) == 4 else (entry[0], r[1], entry)
        # now go through and fix defaults...
        for (entry,dflt) in six.iteritems(plan.defaults):
            if entry not in data: continue
            r = data[entry]
            data[entry] = (r[0], r[1], r[2], dflt)
    elif arg_names is None and arg_abbrevs is None and defaults is None:
        # return the same object if there are no changes to a schema
        return data
    else:
        data = {r[2]:r for r in data}
    # Now we go through and make updates based on the optional arguments
    if arg_names is None: arg_names = {}
    for (entry,arg_name) in six.iteritems(arg_names):
        if entry not in data: continue
        r = data[entry]
        data[entry] = (r[0], arg_name, entry) if len(r) == 3 else (r[0], arg_name, entry, r[3])
    if arg_abbrevs is None: arg_abbrevs = {}
    for (entry,arg_abbrev) in six.iteritems(arg_abbrevs):
        if entry not in data: continue
        r = data[entry]
        data[entry] = (arg_abbrev, r[1], entry) if len(r) == 3 else (arg_abbrev, r[1], entry, r[3])
    if defaults is None: defaults = {}
    for (entry,dflt) in six.iteritems(defaults):
        if entry not in data: continue
        r = data[entry]
        data[entry] = (r[0], r[1], entry, dflt)
    # return the list-ified version of this
    return [tuple(row) for row in six.itervalues(data)]

def argv_parser(instructs,
                arg_names=None, arg_abbrevs=None, defaults=None,
                value_parser=True, filters=None):
    '''
    argv_parser(instructions) is equivalent to CommandLineParser(instructions).
    argv_parser(plan) is equivalent to CommandLineParser(plan_to_argv_instructions(plan)).

    See also help(CommandLineParser) and help(to_argv_schema); all optional arguments accepted by
    these functions are passed along by argv_parser.

    See also help(argv_parse), which is the recommended interface for parsing command-line
    arguments.
    '''
    schema = to_argv_schema(instructs,
                            arg_names=arg_names, arg_abbrevs=arg_abbrevs, defaults=defaults)
    return CommandLineParser(schema, value_parser=value_parser, filters=filters)

def argv_parse(schema, argv, init=None,
               arg_names=None, arg_abbrevs=None, value_parser=True, defaults=None, filters=None):
    '''
    argv_parse(schema, argv) yields the tuple (unparsed_argv, params) where unparsed_argv is a list
      subset of argv that contains only those command line arguments that were not understood by
      the given argument schema and params is a dictionary of parameters as parsed by the given 
      schema. It is equivalent to argv_parser(schema)(argv). See also help(CommandLineParser) for
      information about the instructions format.
    argv_parse(plan, argv) yields a pimms IMap object whose parameters have been initialized from
      the arguments in argv using the given pimms calculation plan as a template for the argument
      schema; see help(to_argv_schema) for information about the way plans are interpreted as argv
      schemas. The plan is initialized with the additional parameters 'argv' and 'argv_parsed'. The
      'argv' parameter contains the command-line arguments in argv that were not interpreted by the
      command-line parser; the 'argv_parsed' parameter contains the parsed command-line parameters.
      To avoid the plan-specific behavior and instead only parse the arguments from a plan, use
      argv_parse(to_argv_schema(plan), argv).

    The following options may be given:
      * init (default: None) specifies that the given dictionary should be merged into either the
        resulting options dictionary (if schema is a schema and not a plan) or into the parameters
        initially provided to the plan (if schema is a plan).
      * arg_names (default: None) may be a dictionary that specifies explicity command-line
        argument names for the plan parameters; plan parameters should be keys and the argument
        names should be values. Any parameter not listed in this option will be interpreted
        according to the above rules. If a parameter is mapped to None then it will not be
        filled from the command-line arguments.
      * arg_abbrevs (default:None) may be a dictionary that is handled identically to that of
        arg_names except that its values must be single letters, which are used for the
        abbreviated flag names.
      * defaults (default: None) may specify the default values for the plan parameters; this
        dictionary overrides the default values of the plan itself.
      * value_parse (default: True) specifies whether the values are interpreted via the
        ast.literal_eval() function. This may be set to False to leave the values as strings or it
        may be set to a function that takes one argument and performs the parsing itself; such a
        function f must obey the syntax `parsed_val = f(string_val)`. The value_parse function is
        only called on arguments that have string values included. Note that by default the
        value_parse function interprets the string '...' as Ellipsis in addition to the typical
        ast.literal_eval() behavior.
      * filters (default: None) optionally specifies a dictionary of filter functions, each of which
        is passed the parsed value of the associated argument. Each filter function f must obey the
        syntax `final_value = f(parsed_value)`. The keys of this dictionary must be the entry names
        of the arguments. Note that filter functions are called on provided default values but the
        value_parse function is not called on these.
    '''
    parser = argv_parser(schema,
                         arg_names=arg_names, arg_abbrevs=arg_abbrevs, defaults=defaults,
                         value_parser=value_parser, filters=filters)
    res = parser(argv)
    if is_plan(schema):
        init = {} if init is None else init
        init = pyr.pmap(init) if not is_pmap(init) else init
        return schema({'argv': tuple(res[0]), 'argv_parsed': pyr.pmap(res[1])}, res[1], init)
    else:
        return res if init is None else (res[0], dict(merge(res[1], init)))

class WorkLog(object):
    '''
    A WorkLog object is a simple print formatter. Given a column width, allows the user to print
    bullet points with customized indentation. Useful for printing progress in long computations
    or workflows.
    '''

    def __init__(self, columns=80, bullet='  * ', stdout=Ellipsis, stderr=Ellipsis):
        if stdout is Ellipsis: stdout = sys.stdout
        if stderr is Ellipsis: stderr = sys.stderr
        self.stdout = stdout
        self.stderr = stderr
        self.columns = columns
        self.bullet = bullet
        self.blanks = ''.join([' ' for x in bullet])
        self.inner_worklog = None
        
    def indent(self, n=1):
        '''
        worklog.indent() yields a duplicate worklog that is indented relative to worklog.
        worklog.indent(x) indents x times.
        '''
        return WorkLog(columns=self.columns,
                       bullet=(self.blanks + self.bullet),
                       stdout=self.stdout, stderr=self.stderr)

    def _write(self, fl, *args):
        if fl is None: return None
        n = self.columns - len(self.blanks)
        bksep = '\n' + self.blanks
        busep = '\n' + self.bullet
        pasep = '\n\n' + self.blanks
        s = self.bullet + busep.join(
            [(bksep.join(textwrap.wrap(arg.strip(), n)) if is_str(arg) else
              pasep.join([bksep.join(textwrap.wrap(s.strip(), n)) for s in arg]))
             for arg in args])
        r = fl.write(s + '\n')
        fl.flush()
        return r
    
    def __call__(self, *args):
        '''
        worlog(a, b, c...) prints each argument as a separate bullet point at the worklog's current
          indentation level.
        
        Each argument may be a string (which is word-wrapped) or a list of strings, each of which is
        word-wrapped and printed as a separate paragraph.
        '''
        return self._write(self.stdout, *args)
    def warn(self, *args):
        '''
        worlog.warn(a, b, c...) prints each argument as a separate bullet point at the worklog's
          current indentation level to the worklog's standard-error.
        
        Each argument may be a string (which is word-wrapped) or a list of strings, each of which is
        word-wrapped and printed as a separate paragraph.
        '''
        return self._write(self.stderr, *args)
    
def worklog(columns=None, bullet='  * ', stdout=Ellipsis, stderr=Ellipsis, verbose=False):
    '''
    worklog() yields a worklog object using sys.stdout and sys.stderr as the outputs.
    worklog(n) yields a worklog that formats output to n columns.

    The following options may be give:
      * bullet (default: '  * ') the bullet-text to print for each bullet.
      * stdout (default: Ellipsis) the file to use in place of stdout; if None, no stdout is
        printed; if Ellipsis, then is set to sys.stdout if verbose is True and None otherwise.
      * stderr (default: Ellipsis) the file to use in place of stderr; if None, no stderr is
        printed; if Ellipsis, then is set to sys.stderr.
      * verbose (default: False) specifies whether to use verbose output. This only has an effect
        if the stdout option is Ellipsis.
    '''
    if columns is None:
        try:              columns = int(os.environ['COLUMNS'])
        except Exception: columns = 80
    if stdout is Ellipsis:
        if verbose: stdout = sys.stdout
        else: stdout = None
    return WorkLog(columns=columns, bullet=bullet, stdout=stdout, stderr=stderr)



####################################################################################################
# pimms/test/__init__.py
#
# This source-code file is part of the pimms library.
#
# The pimms library is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with this program.  If
# not, see <http://www.gnu.org/licenses/>.

'''
The pimms.test test package contains tests for the pimms library as well as examples of the library
usage.
'''

import unittest, math, sys, six, pimms
import numpy as np
import pyrsistent as pyr

if sys.version_info[0] == 3: from   collections import abc as colls
else:                        import collections            as colls

class TestPimms(unittest.TestCase):
    '''
    The TestPimms class defines all the tests for the pimms library.
    '''

    def test_predicates(self):
        '''
        test_predicates ensures that the various pimms functions of the form is_<type>(obj) are
        working properly.
        '''
        # some arbitrary values to use in testing
        qr = pimms.quant([1.5, 3.3, 9.5, 10.4, 6.2, 0.1], 'mm')
        qi = pimms.quant([1, 3, 9, 10, 6, 0], 'seconds')
        mr = np.random.rand(10,3)
        vi = np.random.randint(0, 5000, size=12)
        sr = np.array(10.0)
        si = np.array(2)*pimms.units.deg
        l = [1, 2.0, 'abc']
        lx = [[1,1], [2.0,2.0], [4,7.7]]
        u = u'a unicode string of stuff'
        b = b'a byte string of stuff'
        f0 = lambda:np.linspace(0,100,117)
        f1 = lambda x:x**2 + 1
        d = {'a': 12, 'b':None, 'c':f0, 'd':f1,
             'e':lambda:'some string', 'f':lambda:None}
        pm = pyr.pmap(d)
        lm = pimms.lazy_map(d)
        # a function for testing predicates
        def tpred(p, tvals, fvals):
            for s in tvals: self.assertTrue(p(s))
            for s in fvals: self.assertFalse(p(s))
        # Map types
        tpred(pimms.is_lazy_map, [lm], [qr,qi,mr,vi,sr,si,l,u,b,f0,f1,d,pm])
        tpred(pimms.is_map, [lm,d,pm], [qr,qi,mr,vi,sr,si,l,u,b,f0,f1])
        tpred(pimms.is_pmap, [lm,pm], [qr,qi,mr,vi,sr,si,l,u,b,f0,f1,d])
        # Numpy types require a little attention due to their optional arguments and the
        # complexities of the type relationships
        tpred(pimms.is_nparray, [qr,qi,mr,vi,sr], [l,lx,u,b,f0,f1,d,pm,lm])
        self.assertTrue(pimms.is_nparray(qr, 'real'))
        self.assertTrue(pimms.is_nparray(qr, 'any', 1))
        self.assertFalse(pimms.is_nparray(qr, 'any', 2))
        self.assertFalse(pimms.is_nparray(qi, 'string'))
        self.assertTrue(pimms.is_nparray(qi, ('real','int'), (1,3)))
        self.assertFalse(pimms.is_nparray(qi, ('real','int'), 2))
        self.assertFalse(pimms.is_nparray(qr, ('string','bool','bytes'), (2,3)))
        self.assertTrue(pimms.is_nparray(mr, None, 2))
        self.assertFalse(pimms.is_nparray(mr, None, 1))
        self.assertFalse(pimms.is_nparray(vi, 'int', 2))
        tpred(pimms.is_npscalar, [sr], [qr,qi,mr,vi,l,lx,u,b,f0,f1,d,pm,lm])
        self.assertTrue(pimms.is_npscalar(np.array(12.0), 'real'))
        self.assertFalse(pimms.is_npscalar(np.array(12.0), 'string'))
        self.assertTrue(pimms.is_npscalar(np.array(12.0), ('real','complex')))
        tpred(pimms.is_npmatrix, [mr, pimms.quant(mr, 'm/s')],
              [sr,si,qr,qi,vi,l,lx,u,b,f0,f1,d,pm,lm])
        self.assertTrue(pimms.is_npmatrix(mr, ('int','real','string')))
        self.assertTrue(pimms.is_npmatrix(mr, 'number'))
        self.assertFalse(pimms.is_npmatrix(mr, ('bool','string')))
        tpred(pimms.is_npvector, [qr,qi,vi,vi*pimms.units.mol,qr,qi],
              [sr,si,mr,l,lx,u,b,f0,f1,d,pm,lm])
        self.assertTrue(pimms.is_npvector(vi, 'real'))
        self.assertTrue(pimms.is_npvector(qi, 'int'))
        self.assertFalse(pimms.is_npvector(qr, ('bool','string')))
        self.assertTrue(pimms.is_npvalue('abc', 'string'))
        self.assertTrue(pimms.is_npvalue(u'abc', ('unicode','real')))
        self.assertFalse(pimms.is_npvalue(np.array(5.6), ('unicode','real')))
        self.assertFalse(pimms.is_npvalue(np.array(5.6), ('unicode','bool')))
        self.assertFalse(pimms.is_npvalue(np.array([5.6]), ('unicode','real')))
        # Also the un-nump'ified versions
        tpred(pimms.is_array, [qr,qi,vi,sr,si,mr,qr,qi,l,lx,u,b,f0,f1], [d,pm,lm])
        self.assertTrue(pimms.is_array(qr, 'real'))
        self.assertTrue(pimms.is_array(qr, 'any', 1))
        self.assertTrue(pimms.is_array(qr, 'any', 1))
        self.assertFalse(pimms.is_array(qi, 'string'))
        self.assertTrue(pimms.is_array(qi, ('real','int'), (1,3)))
        self.assertFalse(pimms.is_array(qi, ('real','int'), 2))
        self.assertFalse(pimms.is_array(qr, ('string','bool','bytes'), (2,3)))
        self.assertTrue(pimms.is_array(mr, None, 2))
        self.assertFalse(pimms.is_array(mr, None, 1))
        self.assertFalse(pimms.is_array(vi, 'int', 2))
        self.assertFalse(pimms.is_array(l, 'number', 1))
        self.assertTrue(pimms.is_array(lx, 'any', (1,2)))
        tpred(pimms.is_scalar, [u,b,f0,f1,sr,si], [qr,qi,vi,mr,l,d,pm,lm,lx])
        tpred(pimms.is_int, [vi[0],si,1,10], [u,b,f0,f1,d,pm,lm,sr,qr,mr])
        tpred(pimms.is_real, [vi[0],si,1,10,sr], [4j,u,b,f0,f1,d,pm,lm,lx,mr,qr])
        tpred(pimms.is_complex, [vi[0],si,1,10,sr,4j], [u,b,f0,f1,d,pm,lm,lx,mr,qr])
        tpred(pimms.is_number, [vi[0],si,1,10,sr], [u,b,f0,f1,d,pm,lm,lx,mr,qr])
        tpred(pimms.is_str, ['abc'], [vi,si,1,10,sr,qr,f0,f1,d,pm,lm,lx,mr])
        tpred(pimms.is_class, [str,int], [vi,si,1,10,sr,qr,u,b,f0,f1,d,pm,lm,lx,mr])
        tpred(pimms.is_quantity, [qr,qi,si], [vi,10,sr,u,b,f0,f1,d,pm,lm,lx,mr])
        tpred(pimms.is_unit,
              ('seconds', 's', 'mm', 'deg', pimms.units.seconds, pimms.units.s, pimms.units.mm,
               pimms.units.deg),
              (1, 10.0, np.asarray([10]), None, 'nonunitstring', qr))
        self.assertTrue(pimms.is_nparray(mr, np.inexact))
        self.assertFalse(pimms.is_nparray(vi, np.inexact))
    def test_units(self):
        '''
        test_units ensures that the various pimms functions related to pint integration work
        correctly; these functions include pimms.unit, .mag, .quant, .is_quantity, etc.
        '''
        # make a few pieces of data with types
        x = np.asarray([1.0, 2.0, 3.0, 4.0]) * pimms.units.mm
        y = pimms.quant([2, 4, 6, 8], 'sec')
        for u in [x,y]: self.assertTrue(pimms.is_quantity(u))
        for u in ('abc', 123, 9.0, []): self.assertFalse(pimms.is_quantity(u))
        for u in [x,y]: self.assertFalse(pimms.is_quantity(pimms.mag(u)))
        self.assertTrue(pimms.like_units(pimms.unit(x), pimms.unit('yards')))
        self.assertTrue(pimms.like_units(pimms.unit(y), pimms.unit('minutes')))
        self.assertFalse(pimms.like_units(pimms.unit(y), pimms.unit('mm')))
        z = x / y
        self.assertTrue(pimms.is_vector(x, 'real'))
        self.assertTrue(pimms.is_vector(y, 'real'))
        self.assertFalse(pimms.is_vector(x, 'int'))
        self.assertTrue(pimms.is_vector(y, 'int'))
        self.assertTrue(pimms.is_vector(y, 'float'))
        self.assertTrue(pimms.is_vector(z, 'real'))
        
    def test_lazy_complex(self):
        '''
        test_lazy_complex makes sure that the example in pimms.test.lazy_complex works.
        '''
        from .lazy_complex import LazyComplex

        z1 = LazyComplex(0)
        z2 = LazyComplex((3.0, 4.0))
        z3 = LazyComplex(0, -1)

        # Test the names:
        self.assertEqual(z1.abs, 0)
        self.assertEqual(z1.arg, 0)
        self.assertEqual(z2.abs, 5.0)
        self.assertLess(abs(z2.arg - 0.9272952180016122), 1e-9)
        self.assertEqual(z3.abs, 1)
        self.assertLess(abs(z3.arg - (-0.5 * math.pi)), 1e-9)
        self.assertTrue(z1.is_lazy_complex)
        
    def test_normal_calc(self):
        from .normal_calc import (normal_distribution, pdf)
        
        # instantiate a normal distribution object
        std_norm_dist = normal_distribution(mean=0.0,
                                            standard_deviation=1.0)
        #>> Checking standard_deviation...
        self.assertEqual(std_norm_dist['mean'], 0)
        self.assertEqual(std_norm_dist['standard_deviation'], 1)

        self.assertEqual(
            set(std_norm_dist.keys()),
            set(['mean', 'standard_deviation', 'variance', 'ci95', 'ci99',
                 'inner_normal_distribution_constant',
                 'outer_normal_distribution_constant']))
        
        self.assertEqual(std_norm_dist['variance'], 1)
        #>> Calculating variance...
        
        # this gets cached after being calculated above
        self.assertTrue('variance' in std_norm_dist.efferents)

        self.assertLess(abs(pdf(std_norm_dist, 1.0) - 0.241971), 0.001)
        #>> Calculating outer constant...
        #>> Calculating inner constant...

        # Make a new normal_distribution object similar to the old
        new_norm_dist_1 = std_norm_dist.set(mean=10.0)
        # the standard_deviation check doesn't get rerun because none of
        # it's parameters have changed
        self.assertTrue('variance' in new_norm_dist_1.efferents)
        self.assertEqual(new_norm_dist_1['variance'], 1)
        
        self.assertLess(abs(pdf(new_norm_dist_1, 9.0) - 0.241971), 0.001)

        # Here, the calculations get rerun because the standard_deviation
        # changes
        new_norm_dist_2 = std_norm_dist.set(standard_deviation=2.0)
        #>> Checking standard_deviation...
        self.assertFalse('variance' in new_norm_dist_2.efferents)
        self.assertEqual(new_norm_dist_2['variance'], 4)
        #>> Calculating variance...

    _test_map_af = {'a': 1, 'b': 2, 'c': 3, 'd': 4, 'e': 5, 'f': 6}
    @staticmethod
    def _make_test_counter_map():
        count = []
        def _counter(arg):
            count.append(arg)
            return arg
        def _make_counter(arg):
            return lambda:_counter(arg)
        return (count, {k: _make_counter(v) for (k,v) in six.iteritems(TestPimms._test_map_af)})

    def _assertEquivMaps(self, a, b):
        self.assertTrue(isinstance(a, colls.Mapping))
        self.assertTrue(isinstance(b, colls.Mapping))
        self.assertEqual(len(a), len(b))
        for (f,s) in [(a,b), (b,a)]:
            for (k,v) in six.iteritems(f):
                self.assertTrue(k in s)
                self.assertEqual(v, s[k])
    
    def test_lazy_basics(self):
        # First, test that a lazy map works as a normal map:
        lmap = pimms.lazy_map(TestPimms._test_map_af)
        self._assertEquivMaps(lmap, TestPimms._test_map_af)
        # Now a lazy map
        (c, lmap) = self._make_test_counter_map()
        lmap = pimms.lazy_map(lmap)
        self._assertEquivMaps(lmap, TestPimms._test_map_af)
        self.assertEqual(set(c), set(TestPimms._test_map_af.values()))

    ################################################################################################
    # Test the pimms utilities
    def test_lazy_map(self):
        # We setup a map-testing function that takes a few args and asserts everything is as it
        # should be:
        def _test_lm(lm, m, lazy_ks, norm_ks):
            self.assertTrue(isinstance(lm, pimms.LazyPMap))
            self.assertEqual(len(lm), len(m))
            for k in six.iterkeys(m): self.assertTrue(k in lm)
            # the normal keys should just be normal:
            for k in norm_ks:
                self.assertEqual(m[k], lm[k])
                self.assertTrue(lm.is_normal(k))
                self.assertFalse(lm.is_lazy(k))
                self.assertFalse(lm.is_memoized(k))
            # these tests should not memoize any keys:
            for k in six.iterkeys(lm):
                self.assertTrue(k in m and k in lm)
            for k in lm.iternormal():
                self.assertTrue(k in m and k in norm_ks)
                self.assertEqual(lm[k], m[k])
            self.assertEqual(len(set(lm.itermemoized())), 0)
            self.assertEqual(len(set(lm.iterlazy())), len(lazy_ks))
            # now check the lazy keys (should still all be lazy):
            for k in lazy_ks:
                self.assertFalse(lm.is_normal(k))
                self.assertTrue(lm.is_lazy(k))
                self.assertFalse(lm.is_memoized(k))
                self.assertEqual(m[k], lm[k]) # here the memoization should happen
                self.assertTrue(lm.is_memoized(k))
                self.assertFalse(lm.is_lazy(k))
                self.assertFalse(lm.is_normal(k))
            # everything should be memoized now:
            self.assertEqual(len(set(lm.itermemoized())), len(lazy_ks))
            self.assertEqual(len(set(lm.iterlazy())), 0)
            # it should be an error to delete anything or set anything
            for k in (lm.keys() + ['random_key1', 'random_key2']):
                with self.assertRaises(TypeError): lm[k] = 10
                with self.assertRaises(TypeError): del lm[k]
        # Okay, we can test a few lazy maps now:
        # (1) lazy maps with normal values are just normal maps:
        m = {k:v for (v,k) in enumerate(map(chr, range(ord('a'), ord('z')+1)))}
        lm = pimms.lazy_map(m)
        _test_lm(lm, m, [], m.keys())
        # (2) lazy maps should be build-able...
        norm_ks = set(m.keys())
        lazy_ks = set([])
        d_ord = ord('a') - ord('A')
        def _make_lazy_lambda(v):
            def _fn():
                _make_lazy_lambda.counter += 1
                return v
            return _fn
        _make_lazy_lambda.counter = 0
        for (v,k) in enumerate(map(chr, range(ord('A'), ord('Z') + 1))):
            m[k] = v
            lm = lm.set(k, _make_lazy_lambda(v))
            lazy_ks.add(k)
            if ord(k) % 2 == 0:
                kl = chr(ord(k) + d_ord)
                lm = lm.remove(kl)
                del m[kl]
                norm_ks.remove(kl)
        _test_lm(lm, m, lazy_ks, norm_ks)
        self.assertEqual(_make_lazy_lambda.counter, len(lazy_ks))
        remem = [v for (k,v) in six.iteritems(lm)]
        self.assertEqual(_make_lazy_lambda.counter, len(lazy_ks))

    def test_itable(self):
        '''
        test_itable() tests pimms itable objects and makes sure they work correctly.
        '''
        class nloc:
            lazy_loads = 0
        def _load_lazy():
            nloc.lazy_loads += 1
            return pimms.quant(np.random.rand(10), 'sec')
        dat = pimms.lazy_map({'a': [1,2,3,4,5,6,7,8,9,10],
                              'b': pimms.quant(np.random.rand(10), 'mm'),
                              'c': ['abc','def','ghi','jkl','mno','pqr','stu','vwx','yz!','!!!'],
                              'd': _load_lazy})
        tbl = pimms.itable(dat)
        # make sure the data is the right size
        for k in tbl.keys():
            self.assertTrue(tbl[k].shape == (10,))
        self.assertTrue(tbl.row_count == 10)
        self.assertTrue(len(tbl.rows) == 10)
        self.assertTrue(len(tbl.column_names) == 4)
        self.assertTrue(len(tbl.columns) == 4)
        # Check a few of the entries
        for (i,ki) in zip(np.random.randint(0, tbl.row_count, 50), np.random.randint(0, 4, 50)):
            ki = tbl.column_names[ki]
            self.assertTrue(tbl.rows[i][ki] == tbl[ki][i])
        self.assertTrue(nloc.lazy_loads == 1)
        # see if we can discard stuff
        self.assertTrue('a' in tbl)
        self.assertFalse('a' in tbl.discard('a'))
        self.assertFalse('b' in tbl.discard('b'))
        self.assertFalse('c' in tbl.discard('b').discard('c'))
        self.assertTrue('c' in tbl.discard('b').discard('a'))

    def test_persist(self):
        '''
        test_persist() tests pimms persist() function.
        '''
        from .lazy_complex import LazyComplex

        z = LazyComplex((1.0, 2.0))
        self.assertFalse(z.is_persistent())
        self.assertTrue(z.is_transient())
        z.persist()
        self.assertTrue(z.is_persistent())
        self.assertFalse(z.is_transient())
        
        z = LazyComplex((1.0, 2.0))
        self.assertFalse(z.is_persistent())
        self.assertTrue(z.is_transient())
        zp = pimms.persist(z)
        self.assertTrue(zp.is_persistent())
        self.assertFalse(zp.is_transient())
        self.assertFalse(z.is_persistent())
        self.assertTrue(z.is_transient())

        m0 = {'a': [1,2,3], 'b': (2,3,4),
              'c': {'d':'abc', 'e':set(['def','ghi']), 'f':frozenset([10,11,12])},
              'z': z, 'zp': zp,
              'q': (1,2,[3,4]),
              't': pimms.itable({'c1': range(10), 'c2': range(1,11), 'c3': range(2,12)})}
        m = pimms.persist(m0)
        self.assertIs(m['b'], m0['b'])
        self.assertIsNot(m['a'], m0['a'])
        self.assertTrue(all(ai == bi for (ai,bi) in zip(m['a'], m0['a'])))
        self.assertTrue(pimms.is_pmap(m['c']))
        self.assertIs(m['c']['d'], m0['c']['d'])
        self.assertTrue(isinstance(m['c']['e'], frozenset))
        self.assertTrue(isinstance(m['c']['f'], frozenset))
        self.assertTrue(all(ai == bi for (ai,bi) in zip(m['c']['f'], m0['c']['f'])))
        self.assertTrue(m['z'].is_persistent())
        self.assertIs(m['zp'], m0['zp'])
        self.assertIs(m['q'], m0['q'])
        self.assertIs(m['q'][2], m0['q'][2])
        self.assertTrue(pimms.is_itable(m['t']))
        self.assertTrue(m['t'].is_persistent())
        m = pimms.persist(m0, depth=1)
        self.assertIs(m['b'], m0['b'])
        self.assertIsNot(m['a'], m0['a'])
        self.assertTrue(all(ai == bi for (ai,bi) in zip(m['a'], m0['a'])))
        self.assertTrue(pimms.is_pmap(m['c']))
        self.assertIs(m['c']['d'], m0['c']['d'])
        self.assertTrue(isinstance(m['c']['e'], set))
        self.assertTrue(isinstance(m['c']['f'], frozenset))
        self.assertTrue(all(ai == bi for (ai,bi) in zip(m['c']['f'], m0['c']['f'])))
        self.assertTrue(m['z'].is_persistent())
        self.assertIs(m['zp'], m0['zp'])
        self.assertIs(m['q'], m0['q'])
        self.assertIs(m['q'][2], m0['q'][2])
        self.assertTrue(pimms.is_itable(m['t']))
        self.assertTrue(m['t'].is_persistent())

    def test_cmdline(self):
        '''
        test_cmdline() ensures that the command-line functions such as argv_parse() are working
          correctly and work correctly with calculation plans.
        '''
        from .normal_calc import normal_distribution

        # command-line parsing works with all kinds of string literals:
        schema = [('a', 'arg-a', 'a',  0),
                  ('b', 'arg-b', 'b', ''),
                  ('c', 'arg-c', 'c', {}),
                  ('d', 'arg-d', 'd', []),
                  ('e', 'arg-e', 'e', True),
                  ('f', 'arg-f', 'f', False),
                  ('g', 'arg-g', 'g', Ellipsis),
                  ('h', 'arg-h', 'h', 'test')]
        a = pimms.argv_parse(schema,
                             ['-a1.5', '--arg-b=string arg', '-c', '{10:"str"}', '-d[1,2,3]',
                              '--arg-e', '-g...'])
        (leftover, a) = a
        self.assertEqual(a['a'], 1.5)
        self.assertEqual(a['b'], 'string arg')
        self.assertEqual(a['c'], {10:'str'})
        self.assertEqual(a['d'], [1,2,3])
        self.assertEqual(a['e'], False)
        self.assertEqual(a['f'], False)
        self.assertEqual(a['g'], Ellipsis)
        self.assertEqual(a['h'], 'test')

        # test use with plans:
        dflts = {'mean':0.0, 'standard_deviation':1.0}
        a = pimms.argv_parse(normal_distribution,
                             ['-m4.5', '--standard-deviation=6.3', 'leftover'],
                             defaults=dflts)
        self.assertTrue(pimms.is_imap(a))
        self.assertEqual(a['argv'], ('leftover',))
        self.assertEqual(a['argv_parsed']['mean'], 4.5)
        self.assertEqual(a['mean'], 4.5)
        self.assertEqual(a['standard_deviation'], 6.3)
        a = pimms.argv_parse(normal_distribution,
                             ['--standard-deviation=6.3', 'leftover'],
                             defaults=dflts)
        self.assertTrue(pimms.is_imap(a))
        self.assertEqual(a['argv'], ('leftover',))
        self.assertEqual(a['argv_parsed']['mean'], 0.0)
        self.assertEqual(a['mean'], 0.0)
        self.assertEqual(a['standard_deviation'], 6.3)

if __name__ == '__main__':
    unittest.main()



####################################################################################################
# pimms/test/lazy_complex.py
#
# This source-code file is part of the pimms library.
#
# The pimms library is free software: you can redistribute it and/or modify it under the terms of
# the GNU General Public License as published by the Free Software Foundation, either version 3 of
# the License, or (at your option) any later version.
# 
# This program is distributed in the hope that it will be useful, but WITHOUT ANY WARRANTY; without
# even the implied warranty of  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
# 
# You should have received a copy of the GNU General Public License along with this program.  If
# not, see <http://www.gnu.org/licenses/>.

import pimms
import math

@pimms.immutable
class LazyComplex(object):

    def __init__(self, re, im=None):
        if isinstance(re, tuple) and im is None:
            (re, im) = re
        self.re = re
        self.im = im
    
    @pimms.param
    def re(r):
        return float(r)
    @pimms.option(0.0)
    def im(r):
        return float(r) if r is not None else 0.0

    @pimms.value
    def is_lazy_complex():
        return True
    @pimms.value
    def abs(re, im):
        return math.sqrt(re*re + im*im)
    @pimms.value
    def arg(re, im):
        return math.atan2(im, re)

    def __repr__(self):
        return '(%.3f exp %.3fi)' % (self.abs, self.arg)



# Usage example: implementing a normal-distribution object
import numpy as np
import pimms

# functions can now return the data itself instead of a dict
# if there is only one output, or, if there are more than one,
# they can return a tuple with the data in the same order
# as in the output values
@pimms.calc('variance')
def variance_from_stddev(standard_deviation):
    return standard_deviation ** 2

@pimms.calc('ci95', 'ci99')
def confidence_intervals(mean, standard_deviation):
    ci95 = tuple(q * standard_deviation + mean for q in [-1.96, 1.96])
    ci99 = tuple(q * standard_deviation + mean for q in [-2.58, 2.58])
    return (ci95, ci99)
    # or return {'ci95': ci95, 'ci99': ci99}

# calc functions that are decorated with None explicitly are
# always called whenever any of their dependent parameters are
# updated or when the new dictionary is created; this lets them
# act as checks or tests or requirements; they should not
# return anything
@pimms.calc(None)
def check_normal_distribution_inputs(standard_deviation):
    if standard_deviation <= 0:
        raise ValueError('standard_deviation must be positive')

# If the output name is not provided, it is, by default, the
# name of the function
@pimms.calc
def outer_normal_distribution_constant(standard_deviation):
    return 1.0 / (standard_deviation * np.sqrt(np.pi * 2))
@pimms.calc
def inner_normal_distribution_constant(standard_deviation):
    return -0.5 / standard_deviation

# Now we declare the calc plan: it's just the union of all these
# calc functions, given in any order:
normal_distribution = pimms.plan(
    variance=variance_from_stddev,
    cis=confidence_intervals,
    tests=check_normal_distribution_inputs,
    outer=outer_normal_distribution_constant,
    inner=inner_normal_distribution_constant)

# This function accepts a normal_distribution object and returns
# the probability density at a point
def pdf(nd, x):
    a = nd['outer_normal_distribution_constant']
    b = nd['inner_normal_distribution_constant']
    mu = nd['mean']
    return a * np.exp(b * (x - mu)**2)



